'''
See COPYRIGHT.md for copyright information.

Provides infrastructure for local viewers of GUI applications such as inline XBRL viewers

'''
from arelle.webserver.bottle import Bottle, request, static_file, HTTPResponse
import threading, time, logging, sys, traceback

class LocalViewer:
    noCacheHeaders = {'Cache-Control': 'no-cache, no-store, must-revalidate',
                      'Pragma': 'no-cache',
                      'Expires': '0'}

    def __init__(self, title, staticReportsRoot):
        self.title = title
        self.port = None # viewer unique port
        self.reportsFolders = [staticReportsRoot] # first entry is root of common report files, rest are per-report root
        self.cntlr = None

    def init(self, cntlr, reportsFolder):
        try:
            if self.port is None: # already initialized
                self.cntlr = cntlr

                # find available port
                import socket
                s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
                s.bind(("",0))
                s.listen(1)
                self.port = s.getsockname()[1]
                s.close()

                # start server
                localserver = Bottle()

                localserver.route('/<file:path>', 'GET', self.get)
                localserver.route('<relpath:path>', 'GET', self.get)
                # start local server on the port on a separate thread
                threading.Thread(target=localserver.run,
                                 kwargs=dict(server='cheroot', host='localhost', port=self.port, quiet=True),
                                 daemon=True).start()
                time.sleep(2) # allow other thread to run and start up

            localhost = "http://localhost:{}/{}".format(self.port, len(self.reportsFolders))
            self.reportsFolders.append(reportsFolder)
            self.cntlr.addToLog(_("{}: http://localhost:{}").format(self.title, self.port),
                                messageCode="localViewer:listen",level=logging.DEBUG)
            #cntlr.addToLog("localhost={}".format(localhost), messageCode="localViewer:listen",level=logging.DEBUG)
            return localhost
        except Exception as ex:
            self.cntlr.addToLog(_("{} exception: http://localhost:{} \nException: {} \nTraceback: {}").format(
                self.title, self.port,
                ex, traceback.format_tb(sys.exc_info()[2])), messageCode="localViewer:exception",level=logging.DEBUG)

    def get(self, file=None, relpath=None):
        self.cntlr.addToLog("http://localhost:{}/{}".format(self.port,file), messageCode="localViewer:get",level=logging.DEBUG)
        try:
            return self.getLocalFile(file, relpath, request)
        except HTTPResponse:
            raise # re-raise, such as to support redirects
        except Exception as ex:
            self.cntlr.addToLog(_("{} exception: file: {} \nException: {} \nTraceback: {}").format(
                self.title, file, ex, traceback.format_tb(sys.exc_info()[2])), messageCode="localViewer:exception",level=logging.DEBUG)



from pathlib import PurePath
from tests.integration_tests.validation.conformance_suite_config import ConformanceSuiteConfig

config = ConformanceSuiteConfig(
    file='extensible-enumerations-CONF-2014-10-29/enumerations-index.xml',
    info_url='https://specifications.xbrl.org/work-product-index-extensible-enumerations-extensible-enumerations-1.0.html',
    local_filepath='extensible-enumerations-CONF-2014-10-29.zip',
    name=PurePath(__file__).stem,
    public_download_url='https://www.xbrl.org/2014/extensible-enumerations-CONF-2014-10-29.zip',
)



'''
sphinxMethods processes the Sphinx language in the context of an XBRL DTS and instance.

See COPYRIGHT.md for copyright information.

Sphinx is a Rules Language for XBRL described by a Sphinx 2 Primer
(c) Copyright 2012 CoreFiling, Oxford UK.
Sphinx copyright applies to the Sphinx language, not to this software.
Workiva, Inc. conveys neither rights nor license for the Sphinx language.
'''

from math import exp, fabs, isinf, isnan, log, log10, pow, sqrt
import datetime
from arelle.ModelDtsObject import ModelConcept, ModelRelationship
from arelle.ModelDocument import Type
from arelle.ModelInstanceObject import ModelFact
from arelle.ModelRelationshipSet import ModelRelationshipSet
from arelle.ModelValue import QName, dayTimeDuration, DayTimeDuration
from arelle.ModelXbrl import ModelXbrl
from arelle.ValidateXbrlCalcs import inferredDecimals, inferredPrecision, roundValue
from arelle import XbrlConst, XmlUtil
from numbers import Number
evaluate = None # initialized at end
SphinxException = None
UNBOUND = None
NONE = None


def moduleInit():
    global evaluate, SphinxException, UNBOUND, NONE
    from .SphinxEvaluator import evaluate, SphinxException, UNBOUND, NONE

class Balance(): # fake class for balance type
    pass

class PeriodType(): # fake class for period type
    pass

class Period():
    def __init__(self, start=None, end=None):
        self.start = start
        self.end = end
    @property
    def isForever(self):
        return self.start is None and self.end is None
    @property
    def isInstant(self):
        return self.start is None and self.end is not None
    @property
    def isStartEnd(self):
        return self.start is not None and self.end is not None
    def __repr__(self):
        return "({0},{1})".format(self.start, self.end)
    def __eq__(self, other):
        if isinstance(other, Period):
            return self.start == other.start and self.end == other.end
        return False
    def __ne__(self, other):
        if isinstance(other, Period):
            return self.start != other.start or self.end != other.end
        return False
    def __lt__(self, other):
        if not isinstance(other, Period):
            return False
        if self.isInstant:
            if other.isInstant:
                return self.end < other.end
            elif other.isStartEnd:
                return self.end <= other.start
        elif self.isStartEnd:
            if other.isInstant:
                return self.end < other.end
            elif other.isStartEnd:
                return self.end <= other.start
        return False
    def __le__(self, other):
        if not isinstance(other, Period):
            return False
        if self.isInstant:
            if other.isInstant:
                return self.end <= other.end
            elif other.isStartEnd:
                return self.end <= other.start or self == other
        elif self.isStartEnd:
            if other.isInstant:
                return self.end <= other.end
            elif other.isStartEnd:
                return self.end <= other.start or self == other
        return False
    def __gt__(self, other):
        if not isinstance(other, Period):
            return False
        if self.isInstant:
            if other.isInstant:
                return self.end > other.end
            elif other.isStartEnd:
                return self.start > other.end
        elif self.isStartEnd:
            if other.isInstant:
                return self.end > other.end
            elif other.isStartEnd:
                return self.start > other.end
        return False
    def __ge__(self, other):
        if not isinstance(other, Period):
            return False
        if self.isInstant:
            if other.isInstant:
                return self.end >= other.end
            elif other.isStartEnd:
                return self.start >= other.start or self == other
        elif self.isStartEnd:
            if other.isInstant:
                return self.end >= other.end
            elif other.isStartEnd:
                return self.start >= other.start or self == other
        return False


def hasArg(node, sphinxContext, args, i):
    if i >= len(args):
        raise SphinxException(node, "sphinx.functionArgumentsMismatch",
                              _("Function %(name)s requires %(required)s parameters but %(provided)s are provided"),
                                name=node.name, required=i, provided=len(node.args))

def numericArg(node, sphinxContext, args, i):
    hasArg(node, sphinxContext, args, i)
    arg = args[i]
    if isinstance(arg, Number):
        return arg
    raise SphinxException(node, "sphinx.functionArgumentsMismatch",
                          _("Function %(name)s numeric parameter %(num)s is not a number: %(value)s"),
                          name=node.name, num=i, value=arg)

def numericArgs(node, sphinxContext, args, expectedArgsLen):
    if expectedArgsLen != len(args):
        raise SphinxException(node, "sphinx.functionArgumentsMismatch",
                              _("Function %(name)s requires %(required)s parameters but %(provided)s are provided"),
                              name=node.name, required=expectedArgsLen, provided=len(args))
    numArgs = []
    for i, arg in enumerate(args):
        if i >= expectedArgsLen:
            break
        value = evaluate(arg, sphinxContext, args, value=True)
        if not isinstance(value, Number):
            raise SphinxException(node, "sphinx.functionArgumentsMismatch",
                                  _("Function %(name)s numeric parameters but %(num)s is not numeric: %(value)s"),
                                  num=i, value=value)
            value = 0
        numArgs.append(value)
    for i in range(i, expectedArgsLen):
        numArgs.append(0)
    return numArgs

def strArg(node, sphinxContext, args, i):
    hasArg(node, sphinxContext, args, i)
    arg = args[i]
    if isinstance(arg, str):
        return arg
    raise SphinxException(node, "sphinx.functionArgumentsMismatch",
                          _("Function %(name)s string parameter %(num)s is not a string: %(value)s"),
                          name=node.name, num=i, value=arg)

def strArgs(node, sphinxContext, args, expectedArgsLen):
    if expectedArgsLen != len(args):
        raise SphinxException(node, "sphinx.functionArgumentsMismatch",
                              _("Function %(name)s requires %(required)s parameters but %(provided)s are provided"),
                              name=node.name, required=expectedArgsLen, provided=len(args))
    strArgs = []
    for i, arg in enumerate(args):
        if i >= expectedArgsLen:
            break
        value = evaluate(arg, sphinxContext, value=True)
        if not isinstance(value, str):
            raise SphinxException(node, "sphinx.functionArgumentsMismatch",
                                  _("Function %(name)s string parameters but %(num)s is not numeric: %(value)s"),
                                  name=node.name, num=i, value=value)
            value = 0
        strArgs.append(value)
    for i in range(i, expectedArgsLen):
        strArgs.append(0)
    return strArgs

def factArg(node, sphinxContext, args, i):
    hasArg(node, sphinxContext, args, i)
    fact = evaluate(args[i], sphinxContext, value=False, hsBoundFact=True)
    if isinstance(fact, ModelFact):
        return fact
    raise SphinxException(node, "sphinx.functionArgumentsMismatch",
                          _("Function %(name)s fact parameter %(num)s is not a fact: %(value)s"),
                          name=node.name,
                          num=i, value=fact)

def conceptOrFactArg(node, sphinxContext, args, i):
    hasArg(node, sphinxContext, args, i)
    conceptOrFact = evaluate(args[i], sphinxContext, value=False)
    if isinstance(conceptOrFact, (ModelConcept,ModelFact)):
        return conceptOrFact
    raise SphinxException(node, "sphinx.functionArgumentsMismatch",
                          _("Function %(name)s parameter %(num)s is not a concept or fact: %(value)s"),
                          name=node.name,
                          num=i, value=conceptOrFact)

def conceptArg(node, sphinxContext, args, i):
    hasArg(node, sphinxContext, args, i)
    concept = evaluate(args[i], sphinxContext, value=False)
    if isinstance(concept, ModelConcept):
        return concept
    raise SphinxException(node, "sphinx.functionArgumentsMismatch",
                          _("Function %(name)s fact parameter %(num)s is not a concept: %(value)s"),
                          name=node.name,
                          num=i, value=concept)

def dtsArg(node, sphinxContext, args):
    hasArg(node, sphinxContext, args, 0)
    dts = evaluate(args[0], sphinxContext, value=False)
    if isinstance(dts, ModelXbrl):
        return dts
    raise SphinxException(node, "sphinx.methodArgumentsMismatch",
                          _("Method %(name)s taxonomy parameter is not a DTS: %(value)s"),
                          name=node.name, value=dts)

def networkArg(node, sphinxContext, args):
    hasArg(node, sphinxContext, args, 0)
    network = evaluate(args[0], sphinxContext, value=False)
    if isinstance(network, ModelRelationshipSet):
        return network
    raise SphinxException(node, "sphinx.methodArgumentsMismatch",
                          _("Method %(name)s taxonomy parameter is not a network: %(value)s"),
                          name=node.name, value=network)

def networkConceptArg(node, sphinxContext, args):
    hasArg(node, sphinxContext, args, 1)
    arg = args[1]
    if isinstance(arg, QName):
        arg = network.modelXbrl.qnameConcepts.get(arg, NONE)
    if isinstance(arg, ModelConcept):
        return arg
    raise SphinxException(node, "sphinx.methodArgumentsMismatch",
                          _("Method %(name)s network concept parameter is not a concept: %(value)s"),
                          name=node.name, value=arg)

def relationshipArg(node, sphinxContext, args):
    hasArg(node, sphinxContext, args, 0)
    rel = evaluate(args[0], sphinxContext, value=False)
    if isinstance(rel, ModelRelationship):
        return rel
    raise SphinxException(node, "sphinx.methodArgumentsMismatch",
                          _("Method %(name)s taxonomy parameter is not a relationship: %(value)s"),
                          name=node.name, value=rel)

# numeric functions
NaN = float("NaN")
POSINF = float("INF")
NEGINF = float("-INF")

def _addTimePeriod(node, sphinxContext, args, subtract=False):
    hasArg(node, sphinxContext, args, 0)
    arg = args[0]
    hasArg(node, sphinxContext, args, 1)
    duration = args[1]
    if isinstance(arg, Period) and arg.isInstant and isinstance(duration, DayTimeDuration):
        try:
            return Period(None, (arg.end - duration) if subtract else (arg.end + duration))
        except ValueError:
            raise SphinxException(node, "sphinx.functionArgumentsMismatch",
                                  _("Function %(name)s argument is not an day time duration %(value)s"),
                                  name=node.name, value=duration)
    raise SphinxException(node, "sphinx.functionArgumentsMismatch",
                          _("Function %(name)s argument is not an instant period %(value)s"),
                          name=node.name, value=arg)

def _abs(node, sphinxContext, args):
    args = numericArgs(node, sphinxContext, args, 1)
    return fabs(args[0])

def _balance(node, sphinxContext, args):
    if len(node.args) == 1 and node.args[0] is None:
        return Balance()
    fact = factArg(node, sphinxContext, args, 0)
    return fact.balance

def _booleanFunction(node, sphinxContext, args):
    hasArg(node, sphinxContext, args, 0)
    arg = args[0]
    # >>>> arg can be XML, what does this mean???
    if arg == "true":
        return True
    if args[0] == "false":
        return False
    raise SphinxException(node, "sphinx.functionArgumentsMismatch",
                          _("Function %(name)s fact parameter %(num)s is not \"true\" or \"false\": %(value)s"),
                          name=node.name,
                          num=0, value=arg)

def _concept(node, sphinxContext, args):
    hasArg(node, sphinxContext, args, 0)
    if isinstance(args[0], ModelXbrl):
        dts = dtsArg(node, sphinxContext, args)
        hasArg(node, sphinxContext, args, 1)
        return dts.qnameConcepts.get(args[1], UNBOUND)
    fact = factArg(node, sphinxContext, args, 0)
    return fact.concept


def _concepts(node, sphinxContext, args):
    hasArg(node, sphinxContext, args, 0)
    if isinstance(args[0], ModelXbrl):  # taxonomy concepts
        dts = dtsArg(node, sphinxContext, args)
        return set(concept
                   for concept in dts.qnameConcepts.values()
                   if concept.isItem or concept.isTuple)
    # otherwise must be network concepts
    network = networkArg(node, sphinxContext, args)
    return network.toModelObjects.keys() | network.fromModelObjects.keys()


def _contains(node, sphinxContext, args):
    hasArg(node, sphinxContext, args, 1)
    collection = args[0]
    arg = args[1]
    # period contains
    if isinstance(arg, (Period, datetime.datetime)) and isinstance(collection, Period):
        if collection.isForever:
            return True
        if collection.isDuration:
            if isinstance(arg, datetime):
                return collection.start <= arg <= collection.end
            elif arg.isInstant:
                return collection.start <= arg.end <= collection.end
            elif arg.isDuration:
                return collection.start <= arg.start <= collection.end and collection.start <= arg.end <= collection.end
        if collection.isInstant:
            if isinstance(arg, datetime):
                return collection.end == arg
            if arg.isInstant:
                return collection.end == arg.end
        return False
    # non-period contains
    if isinstance(collection, (tuple, list, set, str)):
        return arg in collection
    raise SphinxException(node, "sphinx.functionArgumentsMismatch",
                          _("Function %(name)s argument is not a collection %(value)s"),
                          name=node.name, value=collection)

def _credit(node, sphinxContext, args):
    hasArg(node, sphinxContext, args, 0)
    if isinstance(args[0], Balance):
        return "credit"
    raise SphinxException(node, "sphinx.functionArgumentsMismatch",
                          _("Function %(name)s argument is not a ::balance %(value)s"),
                          name=node.name, value=args[0])

def _days(node, sphinxContext, args):
    hasArg(node, sphinxContext, args, 0)
    arg = args[0]
    if isinstance(arg, Period):
        if arg.isInstant:
            return 0
        elif arg.isForever:
            return float("inf")
        return (arg.end - arg.start).days
    raise SphinxException(node, "sphinx.functionArgumentsMismatch",
                          _("Function %(name)s argument is not a period %(value)s"),
                          name=node.name, value=arg)

def _debit(node, sphinxContext, args):
    hasArg(node, sphinxContext, args, 0)
    if isinstance(args[0], Balance):
        return "debit"
    raise SphinxException(node, "sphinx.functionArgumentsMismatch",
                          _("Function %(name)s argument is not a ::balance %(value)s"),
                          name=node.name, value=args[0])

def _decimals(node, sphinxContext, args):
    fact = factArg(node, sphinxContext, args, 0)
    return inferredDecimals(fact)

def _dimension(node, sphinxContext, args):
    fact = factArg(node, sphinxContext, args, 0)
    hasArg(node, sphinxContext, args, 1)
    dimQName = args[1]
    context = fact.context
    if context is not None and dimQName in context.qnameDims:
        modelDimension = context.qnameDims[dimQName]
        if modelDimension.isExplicit:
            return modelDimension.memberQname
        else:
            return modelDimension.typedMember
    return NONE

def _dtsDocumentLocations(node, sphinxContext, args):
    dts = dtsArg(node, sphinxContext, args)
    return set(modelDocument.uri
               for modelDocument in dts.urlDocs.values()
               if modelDocument.type in (Type.SCHEMA, Type.LINKBASE))

def _duration(node, sphinxContext, args):
    hasArg(node, sphinxContext, args, 0)
    if isinstance(args[0], PeriodType):
        return "duration"
    raise SphinxException(node, "sphinx.functionArgumentsMismatch",
                          _("Function %(name)s argument is not a ::period-type %(value)s"),
                          name=node.name, value=args[0])

def _durationFunction(node, sphinxContext, args):
    hasArg(node, sphinxContext, args, 1)
    startArg = args[0]
    if isinstance(startArg, str):
        startDateTime = Period(None, XmlUtil.datetimeValue(startArg, none=NONE))
    elif isinstance(startArg, datetime.datetime):
        startDateTime = startArg
    elif isinstance(arg, datetime.date):
        startDateTime = datetime.date(startArg.year, startArg.month, startArg.day)
    endArg = args[1]
    if isinstance(endArg, str):
        endDateTime = Period(None, XmlUtil.datetimeValue(startArg, addOneDay=True, none=NONE))
    elif isinstance(endArg, datetime.datetime):
        endDateTime = endArg
    elif isinstance(endArg, datetime.date):
        endDateTime = datetime.date(endArg.year, endArg.month, endArg.day) + datetime.timedelta(1)
    if (startDateTime and endDateTime):
        return Period(startDateTime, endDateTime)
    raise SphinxException(node, "sphinx.functionArgumentsMismatch",
                          _("Function %(name)s requires two argument that are a date or datetime string or value: %(start)s and %(end)s", ),
                          name=node.name, start=startArg, end=endArg)

def _endDate(node, sphinxContext, args):
    hasArg(node, sphinxContext, args, 0)
    arg = args[0]
    if isinstance(arg, Period):
        if arg.isForever:
            return UNBOUND
        return arg.end
    raise SphinxException(node, "sphinx.functionArgumentsMismatch",
                          _("Function %(name)s argument is not a period %(value)s"),
                          name=node.name, value=arg)

def _endsWith(node, sphinxContext, args):
    value = strArgs(node, sphinxContext, args, 0)
    suffix = strArgs(node, sphinxContext, args, 1)
    return value.endswith(suffix)

def _entityMethod(node, sphinxContext, args):
    fact = factArg(node, sphinxContext, args, 0)
    if fact.context is not None:
        return fact.context.entityIdentifier
    else:
        return UNBOUND

def _entityFunction(node, sphinxContext, args):
    args = strArgs(node, sphinxContext, args, 2)
    return (args[0], args[1])

def _exp(node, sphinxContext, args):
    args = numericArgs(node, sphinxContext, args, 1)
    x = args[0]
    if isnan(x): return NaN
    if x == POSINF: return POSINF
    if x == NEGINF: return 0
    return exp(x)

def _foreverFunction(node, sphinxContext, args):
    return Period(None, None)

def _identifier(node, sphinxContext, args):
    hasArg(node, sphinxContext, args, 0)
    arg = args[0]
    if isinstance(arg, tuple) and len(arg) == 2:
        return arg[1]
    raise SphinxException(node, "sphinx.functionArgumentsMismatch",
                          _("Function %(name)s argument is not an entity identifier %(value)s"),
                          name=node.name, value=arg)

def _hasDecimals(node, sphinxContext, args):
    fact = factArg(node, sphinxContext, args, 0)
    return fact.decimals is not None

def _indexOf(node, sphinxContext, args):
    value = strArgs(node, sphinxContext, args, 0)
    arg = strArgs(node, sphinxContext, args, 1)
    return value.find(arg)

def _hasPrecision(node, sphinxContext, args):
    fact = factArg(node, sphinxContext, args, 0)
    return fact.precision is not None

def _instant(node, sphinxContext, args):
    hasArg(node, sphinxContext, args, 0)
    if isinstance(args[0], PeriodType):
        return "instant"
    raise SphinxException(node, "sphinx.functionArgumentsMismatch",
                          _("Function %(name)s argument is not a ::period-type %(value)s"),
                          name=node.name, value=args[0])

def _instantFunction(node, sphinxContext, args):
    hasArg(node, sphinxContext, args, 0)
    arg = args[0]
    if isinstance(arg, str):
        instDateTime = Period(None, XmlUtil.datetimeValue(arg, addOneDay=True, none=NONE))
        if instDateTime:  # none if date is not valid
            return instDateTime
    elif isinstance(arg, datetime.datetime):
        return Period(None, arg)
    elif isinstance(arg, datetime.date): # must be turned into midnight of the day reported
        return Period(None, datetime.date(arg.year, arg.month, arg.day) + datetime.timedelta(1))
    raise SphinxException(node, "sphinx.functionArgumentsMismatch",
                          _("Function %(name)s argument is not a date or datetime string or value %(value)s"),
                          name=node.name, value=arg)

def _isForever(node, sphinxContext, args):
    hasArg(node, sphinxContext, args, 0)
    arg = args[0]
    if isinstance(arg, Period):
        return arg.isForever
    raise SphinxException(node, "sphinx.functionArgumentsMismatch",
                          _("Function %(name)s argument is not a period %(value)s"),
                          name=node.name, value=arg)

def _isMonetary(node, sphinxContext, args):
    conceptOrFact = conceptOrFactArg(node, sphinxContext, args, 0)
    if isinstance(conceptOrFact, ModelFact):
        return conceptOrFact.concept.isMonetary
    else:
        return conceptOrFact.isMonetary

def _isNumeric(node, sphinxContext, args):
    conceptOrFact = conceptOrFactArg(node, sphinxContext, args, 0)
    return conceptOrFact.isNumeric

def _lastIndexOf(node, sphinxContext, args):
    value = strArgs(node, sphinxContext, args, 0)
    arg = strArgs(node, sphinxContext, args, 1)
    return value.rfind(arg)

def _length(node, sphinxContext, args):
    value = strArgs(node, sphinxContext, args, 0)
    return len(value)

def _lowerCase(node, sphinxContext, args):
    value = strArgs(node, sphinxContext, args, 0)
    return value.lower()

def _ln(node, sphinxContext, args):
    args = numericArgs(node, sphinxContext, args, 1)
    x = args[0]
    if x < 0 or isnan(x): return NaN
    if x == POSINF: return POSINF
    if x == 0: return NEGINF
    return log(x)

def _localPart(node, sphinxContext, args):
    hasArg(node, sphinxContext, args, 0)
    qn = args[0]
    if isinstance(qn, QName):
        return qn.localName
    elif isinstance(qn, str):
        return qn
    raise SphinxException(node, "sphinx.functionArgumentsMismatch",
                          _("Function %(name)s argument is not a QName %(value)s"),
                          name=node.name, value=args[0])

def _log(node, sphinxContext, args):
    args = numericArgs(node, sphinxContext, args, 2)
    x = args[0]
    base = args[1]
    if x < 0 or isnan(x): return NaN
    if x == POSINF: return POSINF
    if x == 0: return POSINF
    if base == 0 or base == POSINF: return 0
    if base == 10: return log10(x)
    return log(x, base)

def _log10(node, sphinxContext, args):
    args = numericArgs(node, sphinxContext, args, 1)
    x = args[0]
    base = args[1]
    if x < 0 or isnan(x): return NaN
    if x == POSINF: return POSINF
    if x == 0: return POSINF
    return log10(x)

def _name(node, sphinxContext, args):
    concept = conceptArg(node, sphinxContext, args, 0)
    return concept.qname

def _numberFunction(node, sphinxContext, args):
    hasArg(node, sphinxContext, args, 0)
    arg = args[0]
    try:
        return float(arg)
    except Exception as ex:
        raise SphinxException(node, "sphinx.functionArgumentsMismatch",
                              _("Function %(name)s argument: %(value)s, error converting to number: %(error)s"),
                              name=node.name, value=arg, error=str(ex))
def _period(node, sphinxContext, args):
    fact = factArg(node, sphinxContext, args, 0)
    context = fact.context
    if context is not None:
        if context.isForeverPeriod:
            return Period()
        if context.isStartEndPeriod:
            return Period(context.startDatetime, context.endDatetime)
        return Period(None, context.instantDatetime)
    return NONE

def _periodType(node, sphinxContext, args):
    if len(node.args) == 1 and node.args[0] is None:
        return PeriodType()
    hasArg(node, sphinxContext, args, 0)
    if isinstance(args[0], Period):
        arg = args[0]
        if arg.isForever:
            return "forever"
        elif arg.isStartEnd:
            return "duration"
        return "instant"
    elif args[0] is NONE: # special case, such as for tuples
        return "none"
    concept = conceptArg(node, sphinxContext, args, 0)
    return concept.periodType

def _power(node, sphinxContext, args):
    args = numericArgs(node, sphinxContext, args, 2)
    x = args[0]
    exp = args[1]
    if isnan(exp) or (isnan(x) and exp != 0) or (isinf(exp) and x in (1, -1)):
        return NaN
    if ((x == POSINF and exp > 0) or
        (x == NEGINF and x > 0 and not x & 1) or
        (exp == POSINF and not -1 <= x <= 1) or
        (x == 0 and exp < 0)):
        return POSINF
    if x == NEGINF and x > 0 and x & 1:
        return NEGINF
    return pow(x, exp)

def _precision(node, sphinxContext, args):
    fact = factArg(node, sphinxContext, args, 0)
    return inferredPrecision(fact)

def _primary(node, sphinxContext, args):
    fact = factArg(node, sphinxContext, args, 0)
    return fact.qname

def _replace(node, sphinxContext, args):
    value = strArgs(node, sphinxContext, args, 0)
    old = strArgs(node, sphinxContext, args, 1)
    new = strArgs(node, sphinxContext, args, 2)
    return value.replace(old, new)

def _roundItem(node, sphinxContext, args):
    fact = factArg(node, sphinxContext, args, 0)
    return roundValue(fact.xValue, decimals=fact.decimals, precision=fact.precision)

def _roundDecimals(node, sphinxContext, args):
    args = numericArgs(node, sphinxContext, args, 2)
    return roundValue(args[0], decmials=args[1])

def _roundPrecision(node, sphinxContext, args):
    args = numericArgs(node, sphinxContext, args, 2)
    return roundValue(args[0], precision=args[1])

def _schemaTypeFunction(node, sphinxContext, args):
    hasArg(node, sphinxContext, args, 0)
    schemaTypeQn = args[0]
    return schemaTypeQn

def _schemaTypeMethod(node, sphinxContext, args):
    concept = conceptArg(node, sphinxContext, args, 0)
    return concept.typeQname

def _signum(node, sphinxContext, args):
    args = numericArgs(node, sphinxContext, args, 1)
    x = args[0]
    if x == 0 or isnan(x): return 0
    if x > 0: return 1
    return -1

def _scenario(node, sphinxContext, args):
    fact = factArg(node, sphinxContext, args, 0)
    return fact.context.segment.nonDimValues("scenario")

def _scheme(node, sphinxContext, args):
    hasArg(node, sphinxContext, args, 0)
    arg = args[0]
    if isinstance(arg, tuple) and len(arg) == 2:
        return arg[0]
    raise SphinxException(node, "sphinx.methodArgumentsMismatch",
                          _("Method %(name)s argument is not an entity identifier %(value)s"),
                          name=node.name, value=arg)

def _segment(node, sphinxContext, args):
    fact = factArg(node, sphinxContext, args, 0)
    return fact.context.segment.nonDimValues("segment")

def _size(node, sphinxContext, args):
    hasArg(node, sphinxContext, args, 0)
    collection = args[0]
    if isinstance(collection, (tuple, list, set)):
        return len(collection)
    raise SphinxException(node, "sphinx.functionArgumentsMismatch",
                          _("Function %(name)s argument is not a collection %(value)s"),
                          name=node.name, value=collection)

def _sqrt(node, sphinxContext, args):
    args = numericArgs(node, sphinxContext, args, 1)
    x = args[0]
    if x < 0 or isnan(x): return NaN
    if x == POSINF: return POSINF
    return sqrt(x)

def _startDate(node, sphinxContext, args):
    hasArg(node, sphinxContext, args, 0)
    arg = args[0]
    if isinstance(arg, Period):
        if arg.isForever:
            return UNBOUND
        if arg.isDuration:
            return arg.start
        return arg.end
    raise SphinxException(node, "sphinx.functionArgumentsMismatch",
                          _("Function %(name)s argument is not a period %(value)s"),
                          name=node.name, value=arg)

def _startsWith(node, sphinxContext, args):
    value = strArgs(node, sphinxContext, args, 0)
    suffix = strArgs(node, sphinxContext, args, 1)
    return value.startswith(suffix)

def _substring(node, sphinxContext, args):
    value = strArgs(node, sphinxContext, args, 0)
    start = strArgs(node, sphinxContext, args, 1)
    end = strArgs(node, sphinxContext, args, 2)
    if isinf(start): start = len(value)
    if isinf(end): end = len(value)
    return value[start:end]

def _subtractTimePeriod(node, sphinxContext, args):
    return _addTimePeriod(node, sphinxContext, args, subtract=True)

def _taxonomy(node, sphinxContext, args):
    return sphinxContext.modelXbrl

def _timePeriodFunction(node, sphinxContext, args):
    duration = strArgs(node, sphinxContext, args, 0)
    return dayTimeDuration(duration)

nowUTC = datetime.datetime.utcnow()
todayUTC = datetime.date(nowUTC.year, nowUTC.month, nowUTC.day)

def _todayUTC(node, sphinxContext, args):
    return todayUTC

def _timeDuration(node, sphinxContext, args):
    duration = strArgs(node, sphinxContext, args, 0)
    return dayTimeDuration(duration)

def toCollection(node, sphinxContext, args, collectionCreator):
    hasArg(node, sphinxContext, args, 0)
    collection = args[0]
    if isinstance(collection, (tuple, list, set)):
        return collectionCreator(collection)
    raise SphinxException(node, "sphinx.methodArgumentsMismatch",
                          _("Method %(name)s argument is not a collection %(value)s"),
                          name=node.name, value=collection)

def _toList(node, sphinxContext, args):
    return toCollection(node, sphinxContext, args, list)

def _toSet(node, sphinxContext, args):
    return toCollection(node, sphinxContext, args, set)

def _trim(node, sphinxContext, args):
    value = strArgs(node, sphinxContext, args, 0)
    return value.strip()

def _tuple(node, sphinxContext, args):
    fact = factArg(node, sphinxContext, args, 0)
    parentModelObject = fact.getparent()
    parentQn = parentModelObject.qname
    if parentQn == XbrlConst.qnXbrliXbrl:
        return NONE
    return parentQn

def _unitMethod(node, sphinxContext, args):
    fact = factArg(node, sphinxContext, args, 0)
    return fact.unit.measures

def _unitFunction(node, sphinxContext, args):
    hasArg(node, sphinxContext, args, 0)
    unitQn = args[0]
    return ([unitQn],[])

def _upperCase(node, sphinxContext, args):
    value = strArgs(node, sphinxContext, args, 0)
    return value.upper()

def _xbrlTypeMethod(node, sphinxContext, args):
    concept = conceptArg(node, sphinxContext, args, 0)
    baseTypeLocalName = concept.baseXbrliType
    if not baseTypeLocalName:
        return NONE
    return QName("{http://www.xbrl.org/2003/instance}xbrli:" + baseTypeLocalName)


# netework functions

def network(node, sphinxContext, args, linkqname=None, linkrole=None, arcqname=None, arcrole=None):
    dts = dtsArg(node, sphinxContext, args)
    return ModelRelationshipSet(dts, arcrole, linkrole=linkrole, linkqname=linkqname, arcqname=arcqname)

def networks(node, sphinxContext, args, linkqname=None, linkrole=None, arcqname=None, arcrole=None):
    networkAllLinkroles = network(node, sphinxContext, args, linkqname, None, arcqname, arcrole)
    linkroles = networkAllLinkroles.linkRoleUris
    return set(network(node, sphinxContext, args, linkqname, linkrole, arcqname, arcrole)
               for linkrole in linkroles)

def _conceptHypercubeAll(node, sphinxContext, args):
    linkrole = strArg(node, sphinxContext, args, 1)
    return network(node, sphinxContext, args, linkrole=linkrole, arcrole=XbrlConst.all)

def _conceptHypercubeAllNetworks(node, sphinxContext, args):
    return networks(node, sphinxContext, args, arcrole=XbrlConst.all)

def _conceptHypercubeNotAll(node, sphinxContext, args):
    linkrole = strArg(node, sphinxContext, args, 1)
    return network(node, sphinxContext, args, linkrole=linkrole, arcrole=XbrlConst.notAll)

def _conceptHypercubeNotAllNetworks(node, sphinxContext, args):
    return networks(node, sphinxContext, args, arcrole=XbrlConst.notAll)

def _dimensionDefault(node, sphinxContext, args):
    linkrole = strArg(node, sphinxContext, args, 1)
    return network(node, sphinxContext, args, linkrole=linkrole, arcrole=XbrlConst.dimensionDefault)

def _dimensionDefaultNetworks(node, sphinxContext, args):
    return networks(node, sphinxContext, args, arcrole=XbrlConst.dimensionDefault)

def _dimensionDomain(node, sphinxContext, args):
    linkrole = strArg(node, sphinxContext, args, 1)
    return network(node, sphinxContext, args, linkrole=linkrole, arcrole=XbrlConst.dimensionDomain)

def _dimensionDomainNetworks(node, sphinxContext, args):
    return networks(node, sphinxContext, args, arcrole=XbrlConst.dimensionDomain)

def _domainMember(node, sphinxContext, args):
    linkrole = strArg(node, sphinxContext, args, 1)
    return network(node, sphinxContext, args, linkrole=linkrole, arcrole=XbrlConst.domainMember)

def _domainMemberNetworks(node, sphinxContext, args):
    return networks(node, sphinxContext, args, arcrole=XbrlConst.domainMember)

def _essenceAlias(node, sphinxContext, args):
    linkrole = strArg(node, sphinxContext, args, 1)
    return network(node, sphinxContext, args, linkrole=linkrole, arcrole=XbrlConst.essenceAlias)

def _essenceAliasNetworks(node, sphinxContext, args):
    return networks(node, sphinxContext, args, arcrole=XbrlConst.essenceAlias)

def _generalSpecial(node, sphinxContext, args):
    linkrole = strArg(node, sphinxContext, args, 1)
    return network(node, sphinxContext, args, linkrole=linkrole, arcrole=XbrlConst.generalSpecial)

def _generalSpecialNetworks(node, sphinxContext, args):
    return networks(node, sphinxContext, args, arcrole=XbrlConst.generalSpecial)

def _genericLink(node, sphinxContext, args):
    linkrole = strArg(node, sphinxContext, args, 1)
    arcrole = strArg(node, sphinxContext, args, 2)
    return network(node, sphinxContext, args, linkrole=linkrole, arcrole=arcrole)

def _genericLinkNetworks(node, sphinxContext, args):
    arcrole = strArg(node, sphinxContext, args, 1)
    return networks(node, sphinxContext, args, arcrole=arcrole)

def _hypercubeDimension(node, sphinxContext, args):
    linkrole = strArg(node, sphinxContext, args, 1)
    return network(node, sphinxContext, args, linkrole=linkrole, arcrole=XbrlConst.hypercubeDimension)

def _hypercubeDimensionNetworks(node, sphinxContext, args):
    return networks(node, sphinxContext, args, arcrole=XbrlConst.hypercubeDimension)

def _network(node, sphinxContext, args, linkqname, linkrole, arcqname, arcrole):
    return network(node, sphinxContext, args, linkqname, linkrole, arcqname, arcrole)

def _networks(node, sphinxContext, args, linkqname, linkrole, arcqname, arcrole):
    return networks(node, sphinxContext, args, linkqname, linkrole, arcqname, arcrole)

def _parentChild(node, sphinxContext, args):
    linkrole = strArg(node, sphinxContext, args, 1)
    return network(node, sphinxContext, args, linkrole=linkrole, arcrole=XbrlConst.parentChild)

def _parentChildNetworks(node, sphinxContext, args):
    return networks(node, sphinxContext, args, arcrole=XbrlConst.parentChild)

def _requiresElement(node, sphinxContext, args):
    linkrole = strArg(node, sphinxContext, args, 1)
    return network(node, sphinxContext, args, linkrole=linkrole, arcrole=XbrlConst.requiresElement)

def _requiresElementNetworks(node, sphinxContext, args):
    return networks(node, sphinxContext, args, arcrole=XbrlConst.requiresElement)

def _similarTuples(node, sphinxContext, args):
    linkrole = strArg(node, sphinxContext, args, 1)
    return network(node, sphinxContext, args, linkrole=linkrole, arcrole=XbrlConst.similarTuples)

def _similarTuplesNetworks(node, sphinxContext, args):
    return networks(node, sphinxContext, args, arcrole=XbrlConst.similarTuples)

def _summationItem(node, sphinxContext, args):
    linkrole = strArg(node, sphinxContext, args, 1)
    return network(node, sphinxContext, args, linkrole=linkrole, arcrole=XbrlConst.summationItem)

def _summationItemNetworks(node, sphinxContext, args):
    return networks(node, sphinxContext, args, arcrole=XbrlConst.summationItem)

# network methods

def _ancestors(node, sphinxContext, args):
    return set(rel.toModelObject
               for rel in _ancestorRelationships(node, sphinxContext, args))

def _ancestorRelationships(node, sphinxContext, args):
    network = networkArg(node, sphinxContext, args)
    concept = networkConceptArg(node, sphinxContext, args)
    def descRels(fromConcept, depth, visited, rels):
        if depth > 0 and concept not in visited:
            visited.add(concept)
            for rel in network.toModelObject(fromConcept):
                rels.add(rel)
                descRels(rel.fromModelObject, depth - 1, visited, rels)
            visited.discard(concept)
    rels = set()
    descRels(concept, numericArg(node, sphinxContext, args, 2), set(), rels)
    return rels

def _arcName(node, sphinxContext, args):
    network = networkArg(node, sphinxContext, args)
    return network.arcqname

def _arcrole(node, sphinxContext, args):
    network = networkArg(node, sphinxContext, args)
    return network.arcrole

def _children(node, sphinxContext, args):
    network = networkArg(node, sphinxContext, args)
    concept = networkConceptArg(node, sphinxContext, args)
    return set(rel.fromModelObject for rel in network.fromModelObject(concept))

def _descendantRelationships(node, sphinxContext, args):
    network = networkArg(node, sphinxContext, args)
    concept = networkConceptArg(node, sphinxContext, args)
    def descRels(fromConcept, depth, visited, rels):
        if depth > 0 and concept not in visited:
            visited.add(concept)
            for rel in network.fromModelObject(fromConcept):
                rels.add(rel)
                descRels(rel.toModelObject, depth - 1, visited, rels)
            visited.discard(concept)
    rels = set()
    descRels(concept, numericArg(node, sphinxContext, args, 2), set(), rels)
    return rels

def _descendants(node, sphinxContext, args):
    return set(rel.toModelObject
               for rel in _descendantRelationships(node, sphinxContext, args))

def _extendedLinkName(node, sphinxContext, args):
    network = networkArg(node, sphinxContext, args)
    return network.linkrole

def _incomingRelationships(node, sphinxContext, args):
    network = networkArg(node, sphinxContext, args)
    concept = networkConceptArg(node, sphinxContext, args)
    return set(network.toModelObject(concept))

def _outgoingRelationships(node, sphinxContext, args):
    network = networkArg(node, sphinxContext, args)
    concept = networkConceptArg(node, sphinxContext, args)
    return set(network.fromModelObject(concept))

def _parents(node, sphinxContext, args):
    network = networkArg(node, sphinxContext, args)
    concept = networkConceptArg(node, sphinxContext, args)
    return set(rel.fromModelObject for rel in network.toModelObject(concept))

def _relationships(node, sphinxContext, args):
    network = networkArg(node, sphinxContext, args)
    return set(network.modelRelationships)

def _linkRole(node, sphinxContext, args):
    network = networkArg(node, sphinxContext, args)
    return network.linkrole

def _sourceConcepts(node, sphinxContext, args):
    network = networkArg(node, sphinxContext, args)
    return network.fromModelObjects().keys()

def _targetConcepts(node, sphinxContext, args):
    network = networkArg(node, sphinxContext, args)
    return network.toModelObjects().keys()

# relationship methods

def _order(node, sphinxContext, args):
    relationship = relationshipArg(node, sphinxContext, args)
    return relationship.order

def _preferredLabel(node, sphinxContext, args):
    relationship = relationshipArg(node, sphinxContext, args)
    return relationship.preferredLabel

def _source(node, sphinxContext, args):
    relationship = relationshipArg(node, sphinxContext, args)
    return relationship.fromModelObject

def _target(node, sphinxContext, args):
    relationship = relationshipArg(node, sphinxContext, args)
    return relationship.toModelObject

def _weight(node, sphinxContext, args):
    relationship = relationshipArg(node, sphinxContext, args)
    return relationship.weight

# aggregative functions

def _all(node, sphinxContext, args):
    return all(args)

def _any(node, sphinxContext, args):
    return any(args)

def _avg(node, sphinxContext, args):
    return sum(args) / len(args)

def _count(node, sphinxContext, args):
    return len(args)

def _exists(node, sphinxContext, args):
    return len(args) > 0

def _first(node, sphinxContext, args):
    if len(args):
        return args[0]
    return UNBOUND

def _list(node, sphinxContext, args):
    return [args]

def _max(node, sphinxContext, args):
    return max(args)

def _median(node, sphinxContext, args):
    if not args: # empty list
        return 0
    args = sorted(args)  # don't sort in place, args may be reused by caller
    middle = int(len(list) / 2)
    if len(list) % 2 == 0:
        return (args[middle - 1] + args[middle]) / 2  # average middle args
    else:
        return args[middle]  # middle item

def _min(node, sphinxContext, args):
    return min(args)

def _missing(node, sphinxContext, args):
    return len(args) == 0

def modes(args):
    counts = {}
    firstAt = {}
    maxcount = 0
    for i, arg in enumerate(args):
        newcount = counts.get(arg, 0) + 1
        counts[arg] = newcount
        if newcount > maxcount:
            maxcount = newcount
        if arg not in firstAt:
            firstAt[arg] = i
    return [(maxcount - count, firstAt[value], value)
            for value, count in counts.items()
            if count == maxcount].sort()

def _mode(node, sphinxContext, args):
    return modes(args)[0][2]

def _modes(node, sphinxContext, args):
    return set(mode[2] for mode in modes(args))

def _stdevp(node, sphinxContext, args):
    return _notImplemented(node, sphinxContext)

def _set(node, sphinxContext, args):
    return set(args)

def _sum(node, sphinxContext, args):
    return sum(args)

def _var(node, sphinxContext, args):
    return _notImplemented(node, sphinxContext)

def _varp(node, sphinxContext, args):
    return _notImplemented(node, sphinxContext)


# miscellaneous methods

def _notImplemented(node, sphinxContext, args):
    sphinxContext.modelXbrl.log("ERROR", "sphinx.functionNotImplemented",
                                _("Function %(name)s is not currently implemented"),
                                sourceFileLine=node.sourceFileLine,
                                name=node.name)
    return NaN

methodImplementation = {
    "abs":          _abs,
    "add-time-period": _addTimePeriod,
    "balance":      _balance,
    "concept":      _concept,
    "concepts":     _concepts,
    "contains":     _contains,
    "credit":       _credit,
    "days":         _days,
    "debit":        _debit,
    "decimals":     _decimals,
    "dimension":    _dimension,
    "dts-document-locations": _dtsDocumentLocations,
    "duration":     _duration,
    "end-date":     _endDate,
    "ends-with":    _endsWith,
    "entity":       _entityMethod,
    "exp":          _exp,
    "has-decimals": _hasDecimals,
    "has-precision":_hasPrecision,
    "identifier":   _identifier,
    "index-of":     _indexOf,
    "instant":      _instant,
    "is-forever":   _isForever,
    "is-monetary":  _isMonetary,
    "is-numeric":   _isNumeric,
    "last-index-of":_lastIndexOf,
    "ln":           _ln,
    "length":       _length,
    "local-part":   _localPart,
    "log":          _log,
    "log10":        _log10,
    "lower-case":   _lowerCase,
    "name":         _name,
    "period":       _period,
    "period-type":  _periodType,
    "power":        _power,
    "precision":    _precision,
    "primary":      _primary,
    "replace":      _replace,
    "round":        _roundItem,
    "round-by-decimals": _roundDecimals,
    "round-by-precision": _roundPrecision,
    "scenario":     _scenario,
    "schema-type":  _schemaTypeMethod,
    "scheme":       _scheme,
    "segment":      _segment,
    "signum":       _signum,
    "size":         _size,
    "sqrt":         _sqrt,
    "start-date":   _startDate,
    "starts-with":  _startsWith,
    "subtract-time-period": _subtractTimePeriod,
    "taxonomy":     _taxonomy,
    "to-list":      _toList,
    "to-set":       _toSet,
    "tuple":        _tuple,
    "unit":         _unitMethod,
    "unknown":      _notImplemented,
    "upper-case":   _upperCase,
    "xbrl-type":    _xbrlTypeMethod,

    #networks
    "concept-hypercube-all":                _conceptHypercubeAll,
    "concept-hypercube-all-networks":       _conceptHypercubeAllNetworks,
    "concept-hypercube-not-all":            _conceptHypercubeNotAll,
    "concept-hypercube-not-all-networks":   _conceptHypercubeNotAllNetworks,
    "dimension-default":                    _dimensionDefault,
    "dimension-default-networks":           _dimensionDefaultNetworks,
    "dimension-domain":                     _dimensionDomain,
    "dimension-domain-networks":            _dimensionDomainNetworks,
    "domain-member":                        _domainMember,
    "domain-member-networks":               _domainMemberNetworks,
    "essence-alias":                        _essenceAlias,
    "essence-alias-networks":               _essenceAliasNetworks,
    "general-special":                      _generalSpecial,
    "general-special-networks":             _generalSpecialNetworks,
    "generic-link":                         _genericLink,
    "generic-link-networks":                _genericLinkNetworks,
    "hypercube-dimension":                  _hypercubeDimension,
    "hypercube-dimension-networks":         _hypercubeDimensionNetworks,
    "network":                              _network,
    "networks":                             _networks,
    "parent-child":                         _parentChild,
    "parent-child-networks":                _parentChildNetworks,
    "requires-element":                     _requiresElement,
    "requires-element-networks":            _requiresElementNetworks,
    "similar-tuples":                       _similarTuples,
    "similar-tuples-networks":              _similarTuplesNetworks,
    "summation-item":                       _summationItem,
    "summation-item-networks":              _summationItemNetworks,

    #network methods
    "ancestors":                            _ancestors,
    "ancestor-relationships":               _ancestorRelationships,
    "arc-name":                             _arcName,
    "arcrole":                              _arcrole,
    "children":                             _children,
    "concepts":                             _concepts,
    "descendant-relationships":             _descendantRelationships,
    "descendants":                          _descendants,
    "extended-link-name":                   _extendedLinkName,
    "incoming-relationships":               _incomingRelationships,
    "outgoing-relationships":               _outgoingRelationships,
    "parents":                              _parents,
    "relationships":                        _relationships,
    "role":                                 _linkRole,
    "source-concepts":                      _sourceConcepts,
    "target-concepts":                      _targetConcepts,

    #relationship methods
    "order":                                _order,
    "preferred-label":                      _preferredLabel,
    "source":                               _source,
    "target":                               _target,
    "weight":                               _weight,
    }

functionImplementation = {
    "boolean":      _booleanFunction,
    "current-date-as-utc": _todayUTC,
    "duration":     _durationFunction,
    "entity":       _entityFunction,
    "forever":      _foreverFunction,
    "instant":      _instantFunction,
    "number":       _numberFunction,
    "schema-type":  _schemaTypeFunction,
    "time-period":  _timePeriodFunction,
    "unit":         _unitFunction,
    }

aggreateFunctionImplementation = {
    "all":          _all,
    "any":          _any,
    "avg":          _avg,
    "count":        _count,
    "exists":       _exists,
    "first":        _first,
    "list":         _list,
    "max":          _max,
    "median":       _median,
    "min":          _min,
    "missing":      _missing,
    "mode":         _mode,
    "modes":        _modes,
    "set":          _set,
    "stdevp":       _stdevp,
    "sum":          _sum,
    "var":          _var,
    "varp":         _varp,
    }

aggreateFunctionAcceptsFactArgs = {
    "all":          False,
    "any":          False,
    "avg":          False,
    "count":        True,
    "exists":       True,
    "first":        True,
    "list":         True,
    "max":          True,
    "median":       False,
    "min":          False,
    "missing":      True,
    "mode":         False,
    "modes":        False,
    "set":          True,
    "stdevp":       False,
    "sum":          False,
    "var":          False,
    "varp":         False,
    }



'''
See COPYRIGHT.md for copyright information.
'''
import os
from arelle import ModelDocument, XbrlConst
from arelle.Version import authorLabel, copyrightLabel

def dislosureSystemTypes(disclosureSystem, *args, **kwargs):
    # return ((disclosure system name, variable name), ...)
    return (("XDC", "XDCplugin"),)

def disclosureSystemConfigURL(disclosureSystem, *args, **kwargs):
    return os.path.join(os.path.dirname(__file__), "config.xml")

def validateXbrlStart(val, parameters=None, *args, **kwargs):
    val.validateXDCplugin = val.validateDisclosureSystem and getattr(val.disclosureSystem, "XDCplugin", False)
    if not (val.validateXDCplugin):
        return


def validateXbrlFinally(val, *args, **kwargs):
    if not (val.validateXDCplugin):
        return

    modelXbrl = val.modelXbrl
    modelDocument = modelXbrl.modelDocument

    _statusMsg = _("validating {0} filing rules").format(val.disclosureSystem.name)
    modelXbrl.profileActivity()
    modelXbrl.modelManager.showStatus(_statusMsg)

    if modelDocument.type in (ModelDocument.Type.INSTANCE, ModelDocument.Type.INLINEXBRL):


        parentChildRels = modelXbrl.relationshipSet(XbrlConst.parentChild)
        referenceRels = modelXbrl.relationshipSet(XbrlConst.conceptReference)

        for qn, facts in modelXbrl.factsByQname.items():
            concept = modelXbrl.qnameConcepts[qn]
            if not parentChildRels.toModelObject(concept):
                modelXbrl.error("XDC:factElementNotPresented",
                                _("Element %(concept)s is used in a fact in the instance, but is not in any presentation relationships."),
                                modelObject=facts, concept=qn)

        requiredConcepts = set(preRel.toModelObject.qname
                               for preRel in parentChildRels.modelRelationships
                               for refRel in referenceRels.fromModelObject(preRel.toModelObject)
                               if refRel.toModelObject.role == "http://www.changhong.com/XDC/role/definitionalAttribute"
                               for refPart in refRel.toModelObject.iterchildren()
                               if refPart.localName == "RequiredInDocument"
                               if refPart.textValue.strip().lower() == "true")

        missingConcepts = requiredConcepts - modelXbrl.factsByQname.keys()
        if missingConcepts:
            modelXbrl.error("XDC:missingRequiredFacts",
                            _("Required facts missing from document: %(concepts)s."),
                            modelObject=modelXbrl, concepts=", ".join(sorted(str(qn) for qn in missingConcepts)))

    modelXbrl.profileActivity(_statusMsg, minTimeToShow=0.0)
    modelXbrl.modelManager.showStatus(None)


__pluginInfo__ = {
    # Do not use _( ) in pluginInfo itself (it is applied later, after loading
    'name': 'Validate XDC',
    'version': '1.0',
    'description': '''XDC Validation.''',
    'license': 'Apache-2',
    'author': authorLabel,
    'copyright': copyrightLabel,
    # classes of mount points (required)
    'DisclosureSystem.Types': dislosureSystemTypes,
    'DisclosureSystem.ConfigURL': disclosureSystemConfigURL,

    'Validate.XBRL.Start': validateXbrlStart,
    'Validate.XBRL.Finally': validateXbrlFinally,
}



'''
EDGAR.py implements an OpenSql database extension for SEC's EDGAR

See COPYRIGHT.md for copyright information.


to use from command line:
   --plugins xbrlDB/ext/EDGAR.py # note this plugin imports xbrlDB plugin

'''
import os, time
from arelle.ModelDocument import Type
from arelle.UrlUtil import authority
from arelle.Version import authorLabel, copyrightLabel

EXT_EDGAR_TABLES = {
                "filing_edgar", "report_edgar",
                "industry_edgar", "industry_edgar_level", "industry_edgar_structure",
                }

countryOfState = {
    "AL": "US","AK": "US","AZ": "US","AR": "US","CA": "US","CO": "US", "CT": "US","DE": "US",
    "FL": "US","GA": "US","HI": "US","ID": "US","IL": "US","IN": "US","IA": "US","KS": "US",
    "KY": "US","LA": "US","ME": "US","MD": "US","MA": "US","MI": "US","MN": "US","MS": "US",
    "MO": "US","MT": "US","NE": "US","NV": "US","NH": "US","NJ": "US","NM": "US","NY": "US",
    "NC": "US","ND": "US","OH": "US","OK": "US","OR": "US","PA": "US","RI": "US","SC": "US",
    "SD": "US","TN": "US","TX": "US","UT": "US","VT": "US","VA": "US","WA": "US","WV": "US",
    "WI": "US","WY": "US","DC": "US","PR": "US","VI": "US","AS": "US","GU": "US","MP": "US",
    "AB": "CA","BC": "CA","MB": "CA","NB": "CA","NL": "CA","NS": "CA","ON": "CA","PE": "CA",
    "QC": "CA","SK": "CA","NT": "CA","NU": "CA","YT": "CA"}


def extEdgarTableDdlFiles(xbrlOpenDb):
    return (EXT_EDGAR_TABLES,
            # may have glob wildcard characters
            [os.path.join("sql", "open", "ext", {"mssql": "edgarMSSqlDB.sql",
                                                 "mysql": "edgarMySqlDB.ddl",
                                                 "sqlite": "edgarSQLiteDB.ddl",
                                                 "orcl": "edgarOracleDB.sql",
                                                 "postgres": "edgarPostgresDB.ddl"}[xbrlOpenDb.product])])

def rssItemGet(rssItem, propertyName):
    if rssItem is not None:
        return getattr(rssItem, propertyName, None)
    return None

def extEdgarMetadata(xbrlOpenDb, entrypoint, rssItem):
    md = xbrlOpenDb.metadata
    # needed for core tables
    md["acceptedTimestamp"] = rssItemGet(rssItem, "acceptanceDatetime") or md.get("acceptance-datetime")
    # needed for ext tables
    md["fiscalYearEnd"] = rssItemGet(rssItem, "fiscalYearEnd") or md.get("fiscal-year-end")
    md["accessionNumber"] = rssItemGet(rssItem, "accessionNumber") or md.get("accession-number") or str(int(time.time()))
    md["filingDate"] = rssItemGet(rssItem, "filingDate") or md.get("filing-date")
    md["authorityHtml_Url"] = rssItemGet(rssItem, "htmlUrl") or md.get("primary-document-url")
    md["entryUrl"] = rssItemGet(rssItem, "url") or md.get("instance-url")
    md["companyName"] = rssItemGet(rssItem, "companyName") or md.get("conformed-name")
    md["zipUrl"] = rssItemGet(rssItem, "enclosureUrl")
    md["fiscalYearFocus"] = md.get("fiscal-year-focus")
    md["fiscalPeriodFocus"] = md.get("fiscal-period-focus")
    md["fiscalYearEnd"] = rssItemGet(rssItem, "fiscalYearEnd") or md.get("fiscal-year-end")
    md["fileNumber"] = rssItemGet(rssItem, "fileNumber") or md.get("file-number")  or str(int(time.time()))
    md["cik"] = rssItemGet(rssItem, "cikNumber") or md.get("cik")
    md["taxNumber"] = md.get("irs-number")
    md["SIC"] = rssItemGet(rssItem, "assignedSic") or md.get("assigned-sic") or -1
    md["filerCategory"] = md.get("filer-category")
    md["publicFloat"] = md.get("public-float")
    md["tradingSymbol"] = md.get("trading-symbol")
    md["stateOfIncorporation"] = md.get("state-of-incorporation")
    md["businessAddressPhone"] = md.get("business-address.phone")
    md["businessAddressStreet1"] = md.get("business-address.street1")
    md["businessAddressStreet2"] = md.get("business-address.street2")
    md["businessAddressCity"] = md.get("business-address.city")
    md["businessAddressState"] = md.get("business-address.state")
    md["businessAddressZip"] = md.get("business-address.zip")
    md["mailAddressStreet1"] = md.get("mail-address.street1")
    md["mailAddressStreet2"] = md.get("mail-address.street2")
    md["mailAddressCity"] = md.get("mail-address.city")
    md["mailAddressState"] = md.get("mail-address.state")
    md["mailAddressZip"] = md.get("mail-address.zip")
    md["formType"] = rssItemGet(rssItem, "formType") or md.get("document-type")
    md["reportId"] = "{}|{}|{}".format(md["cik"], md["formType"].replace("/A",""), md["fiscalYearFocus"], md["fiscalPeriodFocus"])

def extEdgarInitializeBatch(self, rssObject):
    results = self.execute("SELECT fe.accession_number, s.accepted_timestamp "
                           "FROM report r, filing_edgar fe, submission s "
                           "WHERE r.filing_fk = fe.filing_pk AND r.submission_fk = s.submission_pk")
    existingFilings = dict((accessionNumber, timestamp)
                           for accessionNumber, timestamp in results) # timestamp is a string
    for rssItem in rssObject.rssItems:
        if (rssItem.accessionNumber in existingFilings and
            rssItem.acceptanceDatetime == existingFilings[rssItem.accessionNumber]):
            rssItem.skipRssItem = True

def extEdgarExistingFilingPk(xbrlOpenDb):
    accessionNumber = xbrlOpenDb.metadata.get("accessionNumber")
    if accessionNumber:
        results = xbrlOpenDb.execute("SELECT filing_pk FROM filing_edgar WHERE accession_number = '{}'".format(accessionNumber))
        for filingPk, in results:
            return filingPk
    return None

def extEdgarFiling(xbrlOpenDb, now):
    md = xbrlOpenDb.metadata
    table = xbrlOpenDb.getTable('filing_edgar', 'filing_pk',
                          ('filing_pk',
                           'accession_number',
                           'filing_date',
                           'authority_html_url',
                           'entry_url',
                           'entity_name',
                           'zip_url',
                           'fiscal_year',
                           'fiscal_period',
                           'restatement_index',
                           'period_index',
                           'fiscal_year_end',
                           'file_number',
                           'cik',
                           'tax_number',
                           'standard_industry_code',
                           'filer_category',
                           'public_float',
                           'trading_symbol',
                           'legal_state',
                           'phone',
                           'phys_addr1', 'phys_addr2', 'phys_city', 'phys_state', 'phys_zip', 'phys_country',
                           'mail_addr1', 'mail_addr2', 'mail_city', 'mail_state', 'mail_zip', 'mail_country'
                          ),
                          ('filing_pk',),
                          ((xbrlOpenDb.filingPk,
                            md["accessionNumber"],  # NOT NULL
                            md["filingDate"] or now,  # NOT NULL
                            md["authorityHtml_Url"],
                            md["entryUrl"],
                            md["companyName"],
                            md["zipUrl"], # enclsure zip URL if any
                            md["fiscalYearFocus"],
                            md["fiscalPeriodFocus"],
                            None, #'restatement_index',
                            None, #'period_index',
                            md["fiscalYearEnd"],
                            md["fileNumber"],
                            md["cik"],
                            md["taxNumber"],
                            md["SIC"],
                            md["filerCategory"],
                            md["publicFloat"],
                            md["tradingSymbol"],
                            md["stateOfIncorporation"],
                            md["businessAddressPhone"],
                            md["businessAddressStreet1"],
                            md["businessAddressStreet2"],
                            md["businessAddressCity"],
                            md["businessAddressState"],
                            md["businessAddressZip"],
                            countryOfState.get(md["businessAddressState"]),
                            md["mailAddressStreet1"],
                            md["mailAddressStreet2"],
                            md["mailAddressCity"],
                            md["mailAddressState"],
                            md["mailAddressZip"],
                            countryOfState.get(md["mailAddressState"])
                            ),),
                          checkIfExisting=True)

def extEdgarExistingReportPk(xbrlOpenDb):
    md = xbrlOpenDb.metadata
    # find if this exact report is being replaced
    #if md["reportId"]:
    #    results = xbrlOpenDb.execute("SELECT report_pk FROM report WHERE report_id = '{}'".format(md["reportId"]))
    #    for reportPk, in results:
    #        return reportPk
    return None

def extEdgarReport(xbrlOpenDb, now):
    md = xbrlOpenDb.metadata
    table = xbrlOpenDb.getTable('report_edgar', None,
                          ('report_pk',
                           'form_type'
                           ),
                          ('report_pk',),
                          ((xbrlOpenDb.reportPk,
                            md["formType"]
                            ),),
                          checkIfExisting=True)
    if md.get("reportId"):
        xbrlOpenDb.updateTable("report",
                               ("report_id", "is_most_current"),
                               ((md.get("reportId"), False),)
                               )
    xbrlOpenDb.updateTable("report",
                           ("report_pk", "is_most_current"),
                           ((xbrlOpenDb.reportPk, True),)
                           )

def extEdgarReportUpdate(xbrlOpenDb):
    agencySchemaDocId = stdSchemaDocId = None
    for mdlDoc in xbrlOpenDb.modelXbrl.urlDocs.values():
        if mdlDoc in xbrlOpenDb.documentIds:
            for refDoc, ref in mdlDoc.referencesDocument.items():
                if refDoc.inDTS and ref.referenceTypes & {"href", "import", "include"} \
                   and refDoc in xbrlOpenDb.documentIds:
                    if refDoc.type == Type.SCHEMA:
                        nsAuthority = authority(refDoc.targetNamespace, includeScheme=False)
                        nsPath = refDoc.targetNamespace.split('/')
                        if len(nsPath) > 2:
                            if ((nsAuthority in ("fasb.org", "xbrl.us") and nsPath[-2] == "us-gaap") or
                                (nsAuthority == "xbrl.ifrs.org" and nsPath[-1] in ("ifrs", "ifrs-full", "ifrs-smes"))):
                                stdSchemaDocId = xbrlOpenDb.documentIds[refDoc]
                            elif (nsAuthority == "xbrl.sec.gov" and nsPath[-2] == "rr"):
                                agencySchemaDocId = xbrlOpenDb.documentIds[refDoc]
        if agencySchemaDocId or stdSchemaDocId:
            xbrlOpenDb.updateTable("report_edgar",
                                   ("report_pk", "agency_schema_doc_fk", "standard_schema_doc_fk"),
                                   ((xbrlOpenDb.reportPk, agencySchemaDocId, stdSchemaDocId),)
                                   )

__pluginInfo__ = {
    'name': 'xbrlDB Extension for SEC EDGAR',
    'version': '1.0',
    'description': "This plug-in implements additional database fields for U.S. SEC EDGAR.  ",
    'license': 'Apache-2',
    'author': authorLabel,
    'copyright': copyrightLabel,
    'import': ('xbrlDB', ), # import dependent modules
    # classes of mount points (required)
    'xbrlDB.Open.Ext.TableDDLFiles': extEdgarTableDdlFiles,
    'xbrlDB.Open.Ext.Metadata': extEdgarMetadata,
    'xbrlDB.Open.Ext.InitializeBatch': extEdgarInitializeBatch,
    'xbrlDB.Open.Ext.ExistingFilingPk': extEdgarExistingFilingPk,
    'xbrlDB.Open.Ext.ExtFiling': extEdgarFiling,
    'xbrlDB.Open.Ext.ExtReport': extEdgarReport,
    'xbrlDB.Open.Ext.ExistingReportPk': extEdgarExistingReportPk,
    'xbrlDB.Open.Ext.ExtReportUpdate': extEdgarReportUpdate,
}



'''
See COPYRIGHT.md for copyright information.
'''
from collections import defaultdict
from arelle import XbrlConst
from arelle.ModelDtsObject import ModelConcept
import os
emptySet = set()

def checkFilingDimensions(val, drsELRs):

    fromConceptELRs = defaultdict(set)
    hypercubes = set()
    hypercubesInLinkrole = defaultdict(set)
    for ELR in drsELRs:
        domainMemberRelationshipSet = val.modelXbrl.relationshipSet( XbrlConst.domainMember, ELR)

        # check Hypercubes in ELR, accumulate list of primary items
        positiveAxisTableSources = defaultdict(set)
        positiveHypercubes = set()
        primaryItems = set()
        for hasHypercubeArcrole in (XbrlConst.all, XbrlConst.notAll):
            hasHypercubeRelationships = val.modelXbrl.relationshipSet(
                             hasHypercubeArcrole, ELR).fromModelObjects()
            for hasHcRels in hasHypercubeRelationships.values():
                for hasHcRel in hasHcRels:
                    sourceConcept = hasHcRel.fromModelObject
                    primaryItems.add(sourceConcept)
                    hc = hasHcRel.toModelObject
                    hypercubes.add(hc)
                    if hasHypercubeArcrole == XbrlConst.all:
                        positiveHypercubes.add(hc)
                    elif hasHypercubeArcrole == XbrlConst.notAll:
                        if hasHcRel.isClosed:
                            val.modelXbrl.error(("EFM.6.16.06", "GFM.1.08.06"),
                                _("The notAll relationship for table %(hypercube)s must have the closed attribute set to 'false' "
                                  "so that it can achieve its intended purpose of excluding combinations of domain values, on "
                                  "relationship from %(primaryItem)s, in role %(linkrole)s. Please recheck submission."),
                                edgarCode="fs-1606-Not-All-Relationship-Is-Closed",
                                modelObject=hasHcRel, hypercube=hc.qname, linkrole=ELR, primaryItem=sourceConcept.qname)
                        if hc in positiveHypercubes:
                            val.modelXbrl.error(("EFM.6.16.08", "GFM.1.08.08"),
                                _("Element %(hypercube)s is a table that negates itself in %(linkroleDefinition)s for primary item "
                                  "%(primaryItem)s.  Modify the 'all' and 'notAll' relationships that have the element as a target.  "
                                  "Please recheck submission."),
                                edgarCode="fs-1608-Table-Excludes-Itself",
                                modelObject=hasHcRel, hypercube=hc.qname, linkrole=ELR, linkroleDefinition=val.modelXbrl.roleTypeDefinition(ELR), primaryItem=sourceConcept.qname)
                    dimELR = hasHcRel.targetRole
                    dimTargetRequired = (dimELR is not None)
                    if not dimELR:
                        dimELR = ELR
                    hypercubesInLinkrole[dimELR].add(hc) # this is the elr containing the HC-dim relations
                    hcDimRels = val.modelXbrl.relationshipSet(
                             XbrlConst.hypercubeDimension, dimELR).fromModelObject(hc)
                    if dimTargetRequired and len(hcDimRels) == 0:
                        val.modelXbrl.error(("EFM.6.16.09", "GFM.1.08.09"),
                            _("A table is malformed because the %(arcroleURI)s from %(fromConcept)s to %(toConcept)s in link role "
                              "%(linkroleDefinition)s has no consecutive relationships.  "
                              "Remove the targetRole attribute, or provide a consecutive relationship."),
                            edgarCode="fs-1609-Target-Role-With-No-Consecutive-Relationships",
                            modelObject=hasHcRel, hypercube=hc.qname, fromConcept=sourceConcept.qname, toConcept=hc.qname,
                            linkrole=ELR, linkroleDefinition=val.modelXbrl.roleTypeDefinition(ELR),
                            arcroleURI=hasHcRel.arcrole, arcrole=os.path.basename(hasHcRel.arcrole))
                    for hcDimRel in hcDimRels:
                        dim = hcDimRel.toModelObject
                        if isinstance(dim, ModelConcept):
                            domELR = hcDimRel.targetRole
                            domTargetRequired = (domELR is not None)
                            if not domELR:
                                if dim.isExplicitDimension:
                                    domELR = dimELR
                            if hasHypercubeArcrole == XbrlConst.all:
                                positiveAxisTableSources[dim].add(sourceConcept)
                            elif hasHypercubeArcrole == XbrlConst.notAll and \
                                 (dim not in positiveAxisTableSources or \
                                  not commonAncestor(domainMemberRelationshipSet,
                                                  sourceConcept, positiveAxisTableSources[dim])):
                                val.modelXbrl.error(("EFM.6.16.07", "GFM.1.08.08"),
                                    _("Members of axis %(dimension)s are excluded, in role %(linkroleDefinition)s, "
                                      "from primary item %(primaryItem)s but not included in any table.  "
                                      "Please modify the relationships to make it consistent."),
                                    edgarCode="fs-1607-Axis-Excluded-Not-In-Table",
                                    modelObject=hcDimRel, dimension=dim.qname, linkrole=ELR, linkroleDefinition=val.modelXbrl.roleTypeDefinition(ELR), primaryItem=sourceConcept.qname)
                            dimDomRels = val.modelXbrl.relationshipSet(
                                 XbrlConst.dimensionDomain, domELR).fromModelObject(dim)
                            if domTargetRequired and len(dimDomRels) == 0:
                                val.modelXbrl.error(("EFM.6.16.09", "GFM.1.08.09"),
                                    _("A table is malformed because the %(arcroleURI)s from %(fromConcept)s to %(toConcept)s "
                                      "in link role %(linkroleDefinition)s has no consecutive relationships.  "
                                      "Remove the targetRole attribute, or provide a consecutive relationship."),
                                    edgarCode="fs-1609-Target-Role-With-No-Consecutive-Relationships",
                                    modelObject=hcDimRel, dimension=dim.qname, fromConcept=hc.qname, toConcept=dim.qname, linkrole=ELR, linkroleDefinition=val.modelXbrl.roleTypeDefinition(ELR), arcroleURI=hasHcRel.arcrole, arcrole=os.path.basename(hcDimRel.arcrole))
                            # flatten DRS member relationsihps in ELR for undirected cycle detection
                            drsRelsFrom = defaultdict(list)
                            drsRelsTo = defaultdict(list)
                            getDrsRels(val, domELR, dimDomRels, ELR, drsRelsFrom, drsRelsTo)
                            # check for cycles
                            fromConceptELRs[hc].add(dimELR)
                            fromConceptELRs[dim].add(domELR)
                            cycleCausingConcept = undirectedFwdCycle(val, domELR, dimDomRels, ELR, drsRelsFrom, drsRelsTo, fromConceptELRs)
                            if cycleCausingConcept is not None:
                                cycleCausingConcept.append(hcDimRel)
                                val.modelXbrl.error(("EFM.6.16.04", "GFM.1.08.04"),
                                    _("Element %(conceptFrom)s appears in overlapping sets of members in a Directed Relationship "
                                      "Set in role %(linkroleDefinition)s, axis %(dimension)s, path %(path)s.  "
                                      "Remove it as a target of one of its domain-member relationships. Please recheck submission."),
                                    edgarCode="fs-1604-Domain-Is-Tangled",
                                    modelObject=[hc, dim] + [rel for rel in cycleCausingConcept if not isinstance(rel, bool)],
                                    linkrole=ELR, linkroleDefinition=val.modelXbrl.roleTypeDefinition(ELR),
                                    hypercube=hc.qname, dimension=dim.qname, conceptFrom=dim.qname,
                                    path=cyclePath(hc,cycleCausingConcept))
                            fromConceptELRs.clear()
                if hasHypercubeArcrole == XbrlConst.all and len(hasHcRels) > 1:
                    val.modelXbrl.error(("EFM.6.16.05", "GFM.1.08.05"),
                        _("Element %(concept)s is a primary element of more than one Table in relationship group %(linkroleDefinition)s, "
                          "to tables %(hypercubes)s. Please modify the 'all' relationships that have the element as a source so that "
                          "it has only one primary axis."),
                        edgarCode="fs-1605-Primary-Element-Has-Redundant-Tables",
                        modelObject=[sourceConcept] + hasHcRels,
                        hypercubeCount=len(hasHcRels), linkrole=ELR, linkroleDefinition=val.modelXbrl.roleTypeDefinition(ELR),
                        concept=sourceConcept.qname,
                        hypercubes=', '.join(str(r.toModelObject.qname) for r in hasHcRels if isinstance(r.toModelObject, ModelConcept)))

        # check for primary item dimension-member graph undirected cycles
        fromRelationships = domainMemberRelationshipSet.fromModelObjects()
        for relFrom, rels in fromRelationships.items():
            if relFrom in primaryItems:
                drsRelsFrom = defaultdict(list)
                drsRelsTo = defaultdict(list)
                getDrsRels(val, ELR, rels, ELR, drsRelsFrom, drsRelsTo)
                fromConceptELRs[relFrom].add(ELR)
                cycleCausingConcept = undirectedFwdCycle(val, ELR, rels, ELR, drsRelsFrom, drsRelsTo, fromConceptELRs)
                if cycleCausingConcept is not None:
                    val.modelXbrl.error(("EFM.6.16.04", "GFM.1.08.04"),
                        _("Element %(conceptFrom)s appears in overlapping sets of members in a Directed Relationship Set in "
                          "role %(linkroleDefinition)s, from %(conceptFrom)s, path %(path)s.  "
                          "Remove it as a target of one of its domain-member relationships. Please recheck submission."),
                        edgarCode="fs-1604-Domain-Is-Tangled",
                        modelObject=[relFrom] + [rel for rel in cycleCausingConcept if not isinstance(rel, bool)],
                        linkrole=ELR, conceptFrom=relFrom.qname, path=cyclePath(relFrom, cycleCausingConcept),
                        linkroleDefinition=val.modelXbrl.roleTypeDefinition(ELR))
                fromConceptELRs.clear()
            for rel in rels:
                fromMbr = rel.fromModelObject
                toMbr = rel.toModelObject
                toELR = rel.targetRole
                if isinstance(toMbr, ModelConcept) and toELR and len(
                    val.modelXbrl.relationshipSet(
                         XbrlConst.domainMember, toELR).fromModelObject(toMbr)) == 0:
                    val.modelXbrl.error(("EFM.6.16.09", "GFM.1.08.09"),
                        _("A table is malformed because the %(arcroleURI)s from %(fromConcept)s to %(toConcept)s in link role "
                          "%(linkroleDefinition)s has no consecutive relationships.  "
                          "Remove the targetRole attribute, or provide a consecutive relationship."),
                        edgarCode="fs-1609-Target-Role-With-No-Consecutive-Relationships",
                        modelObject=rel, concept=fromMbr.qname, fromConcept=toMbr.qname, toConcept=fromMbr.qname, linkrole=ELR, linkroleDefinition=val.modelXbrl.roleTypeDefinition(ELR), arcroleURI=hasHcRel.arcrole, arcrole=os.path.basename(rel.arcrole))


def getDrsRels(val, fromELR, rels, drsELR, drsRelsFrom, drsRelsTo, fromConcepts=None):
    if not fromConcepts: fromConcepts = set()
    for rel in rels:
        relTo = rel.toModelObject
        if isinstance(relTo, ModelConcept):
            drsRelsFrom[rel.fromModelObject].append(rel)
            drsRelsTo[relTo].append(rel)
            toELR = rel.targetRole
            if not toELR: toELR = fromELR
            if relTo not in fromConcepts:
                fromConcepts.add(relTo)
                domMbrRels = val.modelXbrl.relationshipSet(
                         XbrlConst.domainMember, toELR).fromModelObject(relTo)
                getDrsRels(val, toELR, domMbrRels, drsELR, drsRelsFrom, drsRelsTo, fromConcepts)
                fromConcepts.discard(relTo)
    return False

def undirectedFwdCycle(val, fromELR, rels, drsELR, drsRelsFrom, drsRelsTo, fromConceptELRs, ELRsVisited=None):
    if not ELRsVisited: ELRsVisited = set()
    ELRsVisited.add(fromELR)
    for rel in rels:
        if rel.linkrole == fromELR:
            relTo = rel.toModelObject
            if isinstance(relTo, ModelConcept):
                toELR = rel.targetRole
                if not toELR:
                    toELR = fromELR
                if relTo in fromConceptELRs and toELR in fromConceptELRs[relTo]: #forms a directed cycle
                    return [rel,True]
                fromConceptELRs[relTo].add(toELR)
                if drsRelsFrom:
                    domMbrRels = drsRelsFrom[relTo]
                else:
                    domMbrRels = val.modelXbrl.relationshipSet(
                             XbrlConst.domainMember, toELR).fromModelObject(relTo)
                cycleCausingConcept = undirectedFwdCycle(val, toELR, domMbrRels, drsELR, drsRelsFrom, drsRelsTo, fromConceptELRs, ELRsVisited)
                if cycleCausingConcept is not None:
                    cycleCausingConcept.append(rel)
                    cycleCausingConcept.append(True)
                    return cycleCausingConcept
                fromConceptELRs[relTo].discard(toELR)
                # look for back path in any of the ELRs visited (pass None as ELR)
                cycleCausingConcept = undirectedRevCycle(val, None, relTo, rel, drsELR, drsRelsFrom, drsRelsTo, fromConceptELRs, ELRsVisited)
                if cycleCausingConcept is not None:
                    cycleCausingConcept.append(rel)
                    cycleCausingConcept.append(True)
                    return cycleCausingConcept
    return None

def undirectedRevCycle(val, fromELR, mbrConcept, turnbackRel, drsELR, drsRelsFrom, drsRelsTo, fromConceptELRs, ELRsVisited):
    for arcrole in (XbrlConst.domainMember, XbrlConst.dimensionDomain):
        '''
        for ELR in ELRsVisited if (not fromELR) else (fromELR,):
            for rel in val.modelXbrl.relationshipSet(arcrole, ELR).toModelObject(mbrConcept):
                if not rel.isIdenticalTo(turnbackRel):
                    relFrom = rel.fromModelObject
                    relELR = rel.linkrole
                    if relFrom in fromConcepts and relELR == drsELR:
                        return True
                    if undirectedRevCycle(val, relELR, relFrom, turnbackRel, drsELR, fromConcepts, ELRsVisited):
                        return True
        '''
        if drsRelsTo:
            mbrDomRels = drsRelsTo[mbrConcept]
        else:
            mbrDomRels = val.modelXbrl.relationshipSet(arcrole, None).toModelObject(mbrConcept)
        for rel in mbrDomRels:
            if not rel.isIdenticalTo(turnbackRel):
                relFrom = rel.fromModelObject
                relELR = rel.linkrole
                if relFrom in fromConceptELRs and relELR in fromConceptELRs[relFrom]:
                    return [rel, False] # turnbackRel.toModelObject
                cycleCausingConcept = undirectedRevCycle(val, relELR, relFrom, turnbackRel, drsELR, drsRelsFrom, drsRelsTo, fromConceptELRs, ELRsVisited)
                if cycleCausingConcept is not None:
                    cycleCausingConcept.append(rel)
                    cycleCausingConcept.append(False)
                    return cycleCausingConcept
    return None

def cyclePath(source, cycles):
    isForward = True
    path = []
    for rel in reversed(cycles):
        if isinstance(rel,bool):
            isForward = rel
        else:
            path.append("{0}:{1} {2}".format(rel.modelDocument.basename,
                                             rel.sourceline,
                                             rel.toModelObject.qname if isForward else rel.fromModelObject.qname))
    return str(source.qname) + " " + " - ".join(path)

def commonAncestor(domainMemberRelationshipSet,
                   negSourceConcept, posSourceConcepts):
    negAncestors = ancestorOrSelf(domainMemberRelationshipSet,negSourceConcept)
    for posSourceConcept in posSourceConcepts:
        if len(negAncestors & ancestorOrSelf(domainMemberRelationshipSet,posSourceConcept)):
            return True
    return False

def ancestorOrSelf(domainMemberRelationshipSet,sourceConcept,result=None):
    if not result:
        result = set()
    if not sourceConcept in result:
        result.add(sourceConcept)
        for rels in domainMemberRelationshipSet.toModelObject(sourceConcept):
            ancestorOrSelf(domainMemberRelationshipSet, rels.fromModelObject, result)
    return result



'''
See COPYRIGHT.md for copyright information.

For SEC EDGAR data access see: https://www.sec.gov/os/accessing-edgar-data
e.g., User-Agent: Sample Company Name AdminContact@<sample company domain>.com

'''
from __future__ import annotations
from typing import TYPE_CHECKING, Any, Optional
import os, posixpath, sys, time, calendar, io, json, logging, shutil, zlib
import regex as re
from urllib.parse import quote, unquote
from urllib.error import URLError, HTTPError, ContentTooShortError
from http.client import IncompleteRead
from urllib import request as proxyhandlers

import certifi

try:
    import ssl
except ImportError:
    ssl = None
from arelle.FileSource import SERVER_WEB_CACHE, archiveFilenameParts
from arelle.PluginManager import pluginClassMethods
from arelle.UrlUtil import isHttpUrl
from arelle.Version import __version__

if TYPE_CHECKING:
    from arelle.Cntlr import Cntlr

addServerWebCache = None

DIRECTORY_INDEX_FILE = "!~DirectoryIndex~!"
INF = float("inf")
RETRIEVAL_RETRY_COUNT = 5
HTTP_USER_AGENT = 'Mozilla/5.0 (Arelle/{})'.format(__version__)

def proxyDirFmt(httpProxyTuple):
    if isinstance(httpProxyTuple,(tuple,list)) and len(httpProxyTuple) == 5:
        useOsProxy, urlAddr, urlPort, user, password = httpProxyTuple
        if useOsProxy:
            return None
        elif urlAddr:
            if user and password:
                userPart = "{0}:{1}@".format(user, password)
            else:
                userPart = ""
            if urlPort:
                portPart = ":{0}".format(urlPort)
            else:
                portPart = ""
            return {"http": "http://{0}{1}{2}".format(userPart, urlAddr, portPart) }
            #return {"http": "{0}{1}{2}".format(userPart, urlAddr, portPart) }
        else:
            return {}  # block use of any proxy
    else:
        return None # use system proxy

def proxyTuple(url): # system, none, or http:[user[:passowrd]@]host[:port]
    if url == "none":
        return (False, "", "", "", "")
    elif url == "system":
        return (True, "", "", "", "")
    userpwd, sep, hostport = url.rpartition("://")[2].rpartition("@")
    urlAddr, sep, urlPort = hostport.partition(":")
    user, sep, password = userpwd.partition(":")
    return (False, urlAddr, urlPort, user, password)

def lastModifiedTime(headers):
    if headers:
        headerTimeStamp = headers["last-modified"]
        if headerTimeStamp:
            from email.utils import parsedate
            hdrTime = parsedate(headerTimeStamp)
            if hdrTime:
                return time.mktime(hdrTime)
    return None


class WebCache:

    default_timeout = None

    def __init__(
        self, cntlr: Cntlr,
        httpProxyTuple: tuple[bool, str, str, str, str] | None
    ) -> None:

        self.cntlr = cntlr
        #self.proxies = request.getproxies()
        #self.proxies = {'ftp': 'ftp://63.192.17.1:3128', 'http': 'http://63.192.17.1:3128', 'https': 'https://63.192.17.1:3128'}
        self._timeout = None

        self._noCertificateCheck = False
        self._httpUserAgent = HTTP_USER_AGENT # default user agent for product
        self._httpsRedirect = False
        self.resetProxies(httpProxyTuple)

        self.opener.addheaders = [('User-agent', self.httpUserAgent)]

        #self.opener = WebCacheUrlOpener(cntlr, proxyDirFmt(httpProxyTuple)) # self.proxies)

        if cntlr.isGAE:
            self.cacheDir = SERVER_WEB_CACHE # GAE type servers
            self.encodeFileChars = re.compile(r'[:^]')
        elif sys.platform == "darwin" and "/Application Support/" in cntlr.userAppDir:
            self.cacheDir = cntlr.userAppDir.replace("Application Support","Caches")
            self.encodeFileChars = re.compile(r'[:^]')

        else:  #windows and unix
            self.cacheDir = cntlr.userAppDir + os.sep + "cache"
            if sys.platform.startswith("win"):
                self.encodeFileChars = re.compile(r'[<>:"\\|?*^]')
            else:
                self.encodeFileChars = re.compile(r'[:^]')
        self.decodeFileChars = re.compile(r'\^[0-9]{3}')
        self.workOffline: bool = False
        self._logDownloads = False
        self.maxAgeSeconds = 60.0 * 60.0 * 24.0 * 7.0 # seconds before checking again for file
        if cntlr.hasFileSystem and not cntlr.disablePersistentConfig:
            self.urlCheckJsonFile = cntlr.userAppDir + os.sep + "cachedUrlCheckTimes.json"
            try:
                with io.open(self.urlCheckJsonFile, 'rt', encoding='utf-8') as f:
                    self.cachedUrlCheckTimes = json.load(f)
            except Exception:
                self.cachedUrlCheckTimes = {}
        else:
            self.cachedUrlCheckTimes = {}
        self.cachedUrlCheckTimesModified = False

    @property
    def timeout(self):
        return self._timeout or WebCache.default_timeout

    @timeout.setter
    def timeout(self, seconds):
        self._timeout = seconds

    @property
    def recheck(self):
        days = self.maxAgeSeconds / (60.0 * 60.0 * 24.0)
        if days == INF:
            return "never"
        elif days >= 30:
            return "monthly"
        elif days >= 7:
            return "weekly"
        elif days >=1:
            return "daily"
        elif self.maxAgeSeconds >= 3600.0:
            return "hourly"
        elif self.maxAgeSeconds >= 900.0: # 15 minutes. just intended for testing
            return "quarter-hourly"
        else:
            return "(invalid)"

    @recheck.setter
    def recheck(self, recheckInterval):
        self.maxAgeSeconds = {"daily": 1.0, "weekly": 7.0, "monthly": 30.0, "never": INF,
                              "hourly": 1.0/24.0, "quarter-hourly": 1.0/96.0 # lower numbers for testing purposes
                              }.get(recheckInterval, 7.0) * (60.0 * 60.0 * 24.0)

    @property
    def logDownloads(self):
        return self._logDownloads

    @logDownloads.setter
    def logDownloads(self, _logDownloads):
        self._logDownloads = _logDownloads

    def saveUrlCheckTimes(self) -> None:
        if self.cachedUrlCheckTimesModified and not self.cntlr.disablePersistentConfig:
            with io.open(self.urlCheckJsonFile, 'wt', encoding='utf-8') as f:
                f.write(json.dumps(self.cachedUrlCheckTimes, ensure_ascii=False, indent=0))
            self.cachedUrlCheckTimesModified = False

    @property
    def noCertificateCheck(self):
        return self._noCertificateCheck

    @noCertificateCheck.setter
    def noCertificateCheck(self, check):
        priorValue = self._noCertificateCheck
        self._noCertificateCheck = check
        if priorValue != check:
            self.resetProxies(self._httpProxyTuple)

    @property
    def httpUserAgent(self):
        return self._httpUserAgent

    @httpUserAgent.setter
    def httpUserAgent(self, userAgent):
        if not userAgent: # None or blank sets to default
            userAgent = HTTP_USER_AGENT
        priorValue = self._httpUserAgent
        self._httpUserAgent = userAgent
        if priorValue != userAgent:
            self.resetProxies(self._httpProxyTuple)

    @property
    def httpsRedirect(self):
        return self._httpsRedirect

    @httpsRedirect.setter
    def httpsRedirect(self, value):
        self._httpsRedirect = value

    def resetProxies(self, httpProxyTuple):
        # for ntlm user and password are required
        self.hasNTLM = False
        self._httpProxyTuple = httpProxyTuple # save for resetting in noCertificateCheck setter
        if isinstance(httpProxyTuple,(tuple,list)) and len(httpProxyTuple) == 5:
            useOsProxy, _urlAddr, _urlPort, user, password = httpProxyTuple
            _proxyDirFmt = proxyDirFmt(httpProxyTuple)
            # only try ntlm if user and password are provided because passman is needed
            if user and not useOsProxy:
                for pluginXbrlMethod in pluginClassMethods("Proxy.HTTPAuthenticate"):
                    pluginXbrlMethod(self.cntlr)
                for pluginXbrlMethod in pluginClassMethods("Proxy.HTTPNtlmAuthHandler"):
                    HTTPNtlmAuthHandler = pluginXbrlMethod()
                    if HTTPNtlmAuthHandler is not None:
                        self.hasNTLM = True
                if not self.hasNTLM: # try for python site-packages ntlm
                    try:
                        from ntlm import HTTPNtlmAuthHandler
                        self.hasNTLM = True
                    except ImportError:
                        pass
            if self.hasNTLM:
                pwrdmgr = proxyhandlers.HTTPPasswordMgrWithDefaultRealm()
                pwrdmgr.add_password(None, _proxyDirFmt["http"], user, password)
                self.proxy_handler = proxyhandlers.ProxyHandler({})
                self.proxy_auth_handler = proxyhandlers.ProxyBasicAuthHandler(pwrdmgr)
                self.http_auth_handler = proxyhandlers.HTTPBasicAuthHandler(pwrdmgr)
                self.ntlm_auth_handler = HTTPNtlmAuthHandler.HTTPNtlmAuthHandler(pwrdmgr)
                proxyHandlers = [self.proxy_handler, self.ntlm_auth_handler, self.proxy_auth_handler, self.http_auth_handler]
        if not self.hasNTLM:
            self.proxy_handler = proxyhandlers.ProxyHandler(proxyDirFmt(httpProxyTuple))
            self.proxy_auth_handler = proxyhandlers.ProxyBasicAuthHandler()
            self.http_auth_handler = proxyhandlers.HTTPBasicAuthHandler()
            proxyHandlers = [self.proxy_handler, self.proxy_auth_handler, self.http_auth_handler]
        if ssl:
            # Attempts to load the default CA certificates from the OS.
            context = ssl.create_default_context()
            # Include certifi certificates (Mozilla’s carefully curated
            # collection) for systems with outdated certs and for platforms
            # that we're unable to load certs from (macOS and some Linux
            # distros.)
            context.load_verify_locations(cafile=certifi.where())
            if self.noCertificateCheck:  # this is required in some Akamai environments, such as sec.gov
                context.check_hostname = False
                context.verify_mode = ssl.CERT_NONE
            proxyHandlers.append(proxyhandlers.HTTPSHandler(context=context))
        self.opener = proxyhandlers.build_opener(*proxyHandlers)
        self.opener.addheaders = [
            ('User-Agent', self.httpUserAgent),
            ('Accept-Encoding', 'gzip, deflate')
            ]

        #self.opener.close()
        #self.opener = WebCacheUrlOpener(self.cntlr, proxyDirFmt(httpProxyTuple))

    def normalizeUrl(self, url:Optional[Any], base: Optional[Any] = None) -> Any:
        if url:
            if url.startswith("file://"): url = url[7:]
            elif url.startswith("file:\\"): url = url[6:]
        if url and not (isHttpUrl(url) or os.path.isabs(url)):
            if base is not None and not isHttpUrl(base) and '%' in url:
                url = unquote(url)
            if base:
                if isHttpUrl(base):
                    scheme, sep, path = base.partition("://")
                    normedPath = scheme + sep + posixpath.normpath(os.path.dirname(path) + "/" + url)
                else:
                    if '%' in base:
                        base = unquote(base)
                    normedPath = os.path.normpath(os.path.join(os.path.dirname(base),url))
            else: # includes base == '' (for forcing relative path)
                normedPath = url
            if normedPath.startswith("file://"): normedPath = normedPath[7:]
            elif normedPath.startswith("file:\\"): normedPath = normedPath[6:]

            # no base, not normalized, must be relative to current working directory
            if base is None and not os.path.isabs(url):
                normedPath = os.path.abspath(normedPath)
        else:
            normedPath = url

        if normedPath:
            if isHttpUrl(normedPath):
                scheme, sep, pathpart = normedPath.partition("://")
                pathpart = pathpart.replace('\\','/')
                endingSep = '/' if pathpart[-1] == '/' else ''  # normpath drops ending directory separator
                return scheme + "://" + posixpath.normpath(pathpart) + endingSep
            normedPath = os.path.normpath(normedPath)
            if normedPath.startswith(self.cacheDir):
                normedPath = self.cacheFilepathToUrl(normedPath)
        return normedPath

    def encodeForFilename(self, pathpart):
        return self.encodeFileChars.sub(lambda m: '^{0:03}'.format(ord(m.group(0))), pathpart)

    def urlToCacheFilepath(self, url):
        scheme, sep, path = url.partition("://")
        filepath = [self.cacheDir, scheme]
        pathparts = path.split('/')
        user, sep, server = pathparts[0].partition("@")
        if not sep:
            server = user
            user = None
        host, sep, port = server.partition(':')
        filepath.append(self.encodeForFilename(host))
        if port:
            filepath.append("^port" + port)
        if user:
            filepath.append("^user" + self.encodeForFilename(user) ) # user may have : or other illegal chars
        filepath.extend(self.encodeForFilename(pathpart) for pathpart in pathparts[1:])
        if url.endswith("/"):  # default index file
            filepath.append(DIRECTORY_INDEX_FILE)
        return os.sep.join(filepath)

    def cacheFilepathToUrl(self, cacheFilepath):
        urlparts = cacheFilepath[len(self.cacheDir)+1:].split(os.sep)
        urlparts[0] += ':/'  # add separator between http and file parts, less one '/'
        if len(urlparts) > 2:
            if urlparts[2].startswith("^port"):
                urlparts[1] += ":" + urlparts[2][5:]  # the port number
                del urlparts[2]
            if urlparts[2].startswith("^user"):
                urlparts[1] = urlparts[2][5:] + "@" + urlparts[1]  # the user part
                del urlparts[2]
        if urlparts[-1] == DIRECTORY_INDEX_FILE:
            urlparts[-1] = ""  # restore default index file syntax
        return '/'.join(self.decodeFileChars  # remove cacheDir part
                        .sub(lambda c: chr( int(c.group(0)[1:]) ), # remove ^nnn encoding
                         urlpart) for urlpart in urlparts)

    def getfilename(
            self, url: str | None, base: str | None = None,
            reload: bool = False, checkModifiedTime: bool = False,
            normalize: bool = False, filenameOnly: bool = False) -> str | None:
        if url is None:
            return url
        if base is not None or normalize:
            url = self.normalizeUrl(url, base)
        urlScheme, schemeSep, urlSchemeSpecificPart = url.partition("://")
        if schemeSep and urlScheme in ("http", "https"):
            # is this a mapped archive file contents?
            _archiveFileNameParts = archiveFilenameParts(url)
            if _archiveFileNameParts:
                _archiveFilename = self.getfilename(_archiveFileNameParts[0], reload=reload, checkModifiedTime=checkModifiedTime)
                if _archiveFilename:
                    return os.path.join(_archiveFilename, _archiveFileNameParts[1])
                return None
            # form cache file name (substituting _ for any illegal file characters)
            filepath = self.urlToCacheFilepath(url)
            if self.httpsRedirect:
                if not os.path.exists(filepath):
                    # if enabled, check for missing files in their inverse http/https cache directory
                    redirect = None
                    if url.startswith('http://'):
                        redirect = self.urlToCacheFilepath('https' + url[4:])
                    elif url.startswith('https://'):
                        redirect = self.urlToCacheFilepath('http' + url[5:])
                    if redirect and os.path.exists(redirect):
                        filepath = redirect
            if self.cacheDir == SERVER_WEB_CACHE:
                # server web-cached files are downloaded when opening to prevent excessive memcache api calls
                return filepath
            # quotedUrl has scheme-specific-part quoted except for parameter separators
            quotedUrl = urlScheme + schemeSep + quote(urlSchemeSpecificPart, '/?=&')
            # handle default directory requests
            if filepath.endswith("/"):
                filepath += DIRECTORY_INDEX_FILE
            if os.sep == '\\':
                filepath = filepath.replace('/', '\\')
            if self.workOffline or filenameOnly:
                return filepath
            filepathtmp = filepath + ".tmp"
            fileExt = os.path.splitext(filepath)[1]
            timeNow = time.time()
            timeNowStr = time.strftime('%Y-%m-%dT%H:%M:%S UTC', time.gmtime(timeNow))
            retrievingDueToRecheckInterval = False
            if not reload and os.path.exists(filepath):
                if url in self.cachedUrlCheckTimes and not checkModifiedTime:
                    cachedTime = calendar.timegm(time.strptime(self.cachedUrlCheckTimes[url], '%Y-%m-%dT%H:%M:%S UTC'))
                else:
                    cachedTime = 0
                if timeNow - cachedTime > self.maxAgeSeconds:
                    # weekly check if newer file exists
                    newerOnWeb = False
                    try: # no provision here for proxy authentication!!!
                        remoteFileTime = lastModifiedTime( self.getheaders(quotedUrl) )
                        if remoteFileTime and remoteFileTime > os.path.getmtime(filepath):
                            newerOnWeb = True
                    except:
                        pass # for now, forget about authentication here
                    if not newerOnWeb:
                        # update ctime by copying file and return old file
                        self.cachedUrlCheckTimes[url] = timeNowStr
                        self.cachedUrlCheckTimesModified = True
                        return filepath
                    retrievingDueToRecheckInterval = True
                else:
                    return filepath
            filedir = os.path.dirname(filepath)
            if not os.path.exists(filedir):
                os.makedirs(filedir)
            # Retrieve over HTTP and cache, using rename to avoid collisions
            # self.modelManager.addToLog('web caching: {0}'.format(url))

            # download to a temporary name so it is not left readable corrupted if download fails
            retryCount = 5
            while retryCount > 0:
                try:
                    self.progressUrl = url
                    savedfile, headers, initialBytes = self.retrieve(
                    #savedfile, headers = self.opener.retrieve(
                                      quotedUrl,
                                      filename=filepathtmp,
                                      reporthook=self.reportProgress)

                    # check if this is a real file or a wifi or web logon screen
                    if fileExt in {".xsd", ".xml", ".xbrl"}:
                        if b"<html" in initialBytes:
                            if retrievingDueToRecheckInterval:
                                return self.internetRecheckFailedRecovery(filepath, url,
                                                                          "file contents appear to be an html logon request",
                                                                          timeNowStr)
                            response = None  # found possible logon request
                            if self.cntlr.hasGui:
                                response = self.cntlr.internet_logon(url, quotedUrl,
                                                                     _("Unexpected HTML in {0}").format(url),
                                                                     _("Is this a logon page? If so, click 'yes', else click 'no' if it is the expected XBRL content, or 'cancel' to abort retrieval: \n\n{0}")
                                                                     .format(initialBytes[:1500]))
                            if response == "retry":
                                retryCount -= 1
                                continue
                            elif response != "no":
                                self.cntlr.addToLog(_("Web file appears to be an html logon request, not retrieved: %(URL)s \nContents: \n%(contents)s"),
                                                    messageCode="webCache:invalidRetrieval",
                                                    messageArgs={"URL": url, "contents": initialBytes},
                                                    level=logging.ERROR)
                                return None

                    retryCount = 0
                except (ContentTooShortError, IncompleteRead) as err:
                    if retrievingDueToRecheckInterval:
                        return self.internetRecheckFailedRecovery(filepath, url, err, timeNowStr)
                    if retryCount > 1:
                        self.cntlr.addToLog(_("%(error)s \nunsuccessful retrieval of %(URL)s \n%(retryCount)s retries remaining"),
                                            messageCode="webCache:retryingOperation",
                                            messageArgs={"error": err, "URL": url, "retryCount": retryCount},
                                            level=logging.ERROR)
                        retryCount -= 1
                        continue
                    self.cntlr.addToLog(_("%(error)s \nretrieving %(URL)s"),
                                        messageCode="webCache:contentTooShortError",
                                        messageArgs={"URL": url, "error": err},
                                        level=logging.ERROR)
                    if os.path.exists(filepathtmp):
                        os.remove(filepathtmp)
                    return None
                    # handle file is bad
                except (HTTPError, URLError) as err:
                    try:
                        tryWebAuthentication = False
                        if isinstance(err, HTTPError) and err.code == 401:
                            tryWebAuthentication = True
                            if 'www-authenticate' in err.hdrs:
                                match = re.match('[ \t]*([^ \t]+)[ \t]+realm="([^"]*)"', err.hdrs['www-authenticate'])
                                if match:
                                    scheme, realm = match.groups()
                                    if scheme.lower() == 'basic':
                                        host = os.path.dirname(quotedUrl)
                                        userPwd = self.cntlr.internet_user_password(host, realm)
                                        if isinstance(userPwd,(tuple,list)):
                                            self.http_auth_handler.add_password(realm=realm,uri=host,user=userPwd[0],passwd=userPwd[1])
                                            retryCount -= 1
                                            continue
                                    self.cntlr.addToLog(_("'%(scheme)s' www-authentication for realm '%(realm)s' is required to access %(URL)s\n%(error)s"),
                                                        messageCode="webCache:unsupportedWWWAuthentication",
                                                        messageArgs={"scheme": scheme, "realm": realm, "URL": url, "error": err},
                                                        level=logging.ERROR)
                        elif isinstance(err, HTTPError) and err.code == 407:
                            tryWebAuthentication = True
                            if 'proxy-authenticate' in err.hdrs:
                                match = re.match('[ \t]*([^ \t]+)[ \t]+realm="([^"]*)"', err.hdrs['proxy-authenticate'])
                                if match:
                                    scheme, realm = match.groups()
                                    host = self.proxy_handler.proxies.get('http')
                                    if scheme.lower() == 'basic':
                                        userPwd = self.cntlr.internet_user_password(host, realm)
                                        if isinstance(userPwd,(tuple,list)):
                                            self.proxy_auth_handler.add_password(realm=realm,uri=host,user=userPwd[0],passwd=userPwd[1])
                                            retryCount -= 1
                                            continue
                                    self.cntlr.addToLog(_("'%(scheme)s' proxy-authentication for realm '%(realm)s' is required to access %(URL)s\n%(error)s"),
                                                        messageCode="webCache:unsupportedProxyAuthentication",
                                                        messageArgs={"scheme": scheme, "realm": realm, "URL": url, "error": err},
                                                        level=logging.ERROR)
                        if retrievingDueToRecheckInterval:
                            return self.internetRecheckFailedRecovery(filepath, url, err, timeNowStr)
                        if tryWebAuthentication:
                            # check if single signon is requested (on first retry)
                            if retryCount == RETRIEVAL_RETRY_COUNT:
                                for pluginXbrlMethod in pluginClassMethods("Proxy.HTTPAuthenticate"):
                                    if pluginXbrlMethod(self.cntlr): # true if succeessful single sign on
                                        retryCount -= 1
                                        break
                                if retryCount < RETRIEVAL_RETRY_COUNT:
                                    continue # succeeded
                            # may be a web login authentication request
                            response = None  # found possible logon request
                            if self.cntlr.hasGui:
                                response = self.cntlr.internet_logon(url, quotedUrl,
                                                                     _("HTTP {0} authentication request").format(err.code),
                                                                     _("Is browser-based internet access authentication possible? If so, click 'yes', or 'cancel' to abort retrieval: \n\n{0}")
                                                                     .format(url))
                            if response == "retry":
                                retryCount -= 1
                                continue
                            elif response != "no":
                                self.cntlr.addToLog(_("Web file HTTP 401 (authentication required) response, not retrieved: %(URL)s"),
                                                    messageCode="webCache:authenticationRequired",
                                                    messageArgs={"URL": url},
                                                    level=logging.ERROR)
                                return None

                    except AttributeError:
                        pass
                    if retrievingDueToRecheckInterval:
                        return self.internetRecheckFailedRecovery(filepath, url, err, timeNowStr)
                    self.cntlr.addToLog(_("%(error)s \nretrieving %(URL)s"),
                                        messageCode="webCache:retrievalError",
                                        messageArgs={"error": err.reason if hasattr(err, "reason") else err,
                                                     "URL": url},
                                        level=logging.ERROR)
                    return None

                except Exception as err:
                    if retryCount > 1:
                        self.cntlr.addToLog(_("%(error)s \nunsuccessful retrieval of %(URL)s \n%(retryCount)s retries remaining"),
                                            messageCode="webCache:retryingOperation",
                                            messageArgs={"error": err, "URL": url, "retryCount": retryCount},
                                            level=logging.ERROR)
                        retryCount -= 1
                        continue
                    if retrievingDueToRecheckInterval:
                        return self.internetRecheckFailedRecovery(filepath, url, err, timeNowStr)
                    if self.cntlr.hasGui:
                        self.cntlr.addToLog(_("%(error)s \nunsuccessful retrieval of %(URL)s \nswitching to work offline"),
                                            messageCode="webCache:attemptingOfflineOperation",
                                            messageArgs={"error": err, "URL": url},
                                            level=logging.ERROR)
                        # try working offline
                        self.workOffline = True
                        return filepath
                    else:  # don't switch offline unexpectedly in scripted (batch) operation
                        self.cntlr.addToLog(_("%(error)s \nunsuccessful retrieval of %(URL)s"),
                                            messageCode="webCache:unsuccessfulRetrieval",
                                            messageArgs={"error": err, "URL": url},
                                            level=logging.ERROR)
                        if os.path.exists(filepathtmp):
                            os.remove(filepathtmp)
                        return None

                # rename temporarily named downloaded file to desired name
                if os.path.exists(filepath):
                    try:
                        if os.path.isfile(filepath) or os.path.islink(filepath):
                            os.remove(filepath)
                        elif os.path.isdir(filepath):
                            shutil.rmtree(filepath)
                    except Exception as err:
                        self.cntlr.addToLog(_("%(error)s \nUnsuccessful removal of prior file %(filepath)s \nPlease remove with file manager."),
                                            messageCode="webCache:cachedPriorFileLocked",
                                            messageArgs={"error": err, "filepath": filepath},
                                            level=logging.ERROR)
                try:
                    os.rename(filepathtmp, filepath)
                    if self._logDownloads:
                        self.cntlr.addToLog(_("Downloaded %(URL)s"),
                                            messageCode="webCache:download",
                                            messageArgs={"URL": url, "filepath": filepath},
                                            level=logging.INFO)
                except Exception as err:
                    self.cntlr.addToLog(_("%(error)s \nUnsuccessful renaming of downloaded file to active file %(filepath)s \nPlease remove with file manager."),
                                        messageCode="webCache:cacheDownloadRenamingError",
                                        messageArgs={"error": err, "filepath": filepath},
                                        level=logging.ERROR)
                webFileTime = lastModifiedTime(headers)
                if webFileTime: # set mtime to web mtime
                    os.utime(filepath,(webFileTime,webFileTime))
                self.cachedUrlCheckTimes[url] = timeNowStr
                self.cachedUrlCheckTimesModified = True
                return filepath

        if url.startswith("file://"): url = url[7:]
        elif url.startswith("file:\\"): url = url[6:]
        if os.sep == '\\':
            url = url.replace('/', '\\')
        return url

    def internetRecheckFailedRecovery(self, filepath, url, err, timeNowStr):
        self.cntlr.addToLog(_("During refresh of web file ignoring error: %(error)s for %(URL)s"),
                            messageCode="webCache:unableToRefreshFile",
                            messageArgs={"URL": url, "error": err},
                            level=logging.INFO)
        # skip this checking cycle, act as if retrieval was ok
        self.cachedUrlCheckTimes[url] = timeNowStr
        self.cachedUrlCheckTimesModified = True
        return filepath

    def reportProgress(self, blockCount, blockSize, totalSize):
        if totalSize > 0:
            self.cntlr.showStatus(_("web caching {0}: {1:.0f} of {2:.0f} KB").format(
                    self.progressUrl,
                    blockCount * blockSize / 1024,
                    totalSize / 1024))
        else:
            self.cntlr.showStatus(_("web caching {0}: {1:.0f} KB").format(
                    self.progressUrl,
                    blockCount * blockSize / 1024))

    def clear(self):
        for cachedProtocol in ("http", "https"):
            cachedProtocolDir = os.path.join(self.cacheDir, cachedProtocol)
            if os.path.exists(cachedProtocolDir):
                shutil.rmtree(cachedProtocolDir, True)

    def getheaders(self, url):
        if url and isHttpUrl(url):
            try:
                fp = self.opener.open(url, timeout=self.timeout)
                headers = fp.info()
                fp.close()
                return headers
            except Exception:
                pass
        return {}

    def geturl(self, url):  # get the url that the argument url redirects or resolves to
        if url and isHttpUrl(url):
            try:
                fp = self.opener.open(url, timeout=self.timeout)
                actualurl = fp.geturl()
                fp.close()
                return actualurl
            except Exception:
                pass
        return None

    def retrieve(self, url, filename=None, filestream=None, reporthook=None, data=None):
        # return filename, headers (in dict), initial file bytes (to detect logon requests)
        headers = None
        initialBytes = b''
        fp = self.opener.open(url, data, timeout=self.timeout)
        try:
            headers = fp.info()
            if filename:
                tfp = open(filename, 'wb')
            elif filestream:
                tfp = filestream
            try:
                result = filename, headers
                bs = 1024*8
                size = -1
                read = 0
                blocknum = 0
                if reporthook:
                    if "content-length" in headers:
                        size = int(headers["Content-Length"])
                    reporthook(blocknum, bs, size)
                isGzipped = "gzip" in headers.get("content-encoding", "")
                if isGzipped:
                    decompressor = zlib.decompressobj(16+zlib.MAX_WBITS) #this magic number can be inferred from the structure of a gzip file
                while 1:
                    block = fp.read(bs)
                    if not block:
                        break
                    if isGzipped:
                        block = decompressor.decompress(block)
                    read += len(block)
                    tfp.write(block)
                    if blocknum == 0:
                        initialBytes = block
                    blocknum += 1
                    if reporthook:
                        reporthook(blocknum, bs, size)
            finally:
                if filename:
                    tfp.close()
        finally:
            if fp:
                fp.close()
        # raise exception if actual size does not match content-length header
        if size >= 0 and read < size:
            raise ContentTooShortError(
                _("retrieval incomplete: got only %i out of %i bytes")
                % (read, size), result)

        if filestream:
            tfp.seek(0)
        return filename, headers, initialBytes

'''
class WebCacheUrlOpener(request.FancyURLopener):
    def __init__(self, cntlr, proxies=None):
        self.cntlr = cntlr
        super(WebCacheUrlOpener, self).__init__(proxies)
        self.version = 'Mozilla/5.0'

    def http_error_401(self, url, fp, errcode, errmsg, headers, data=None, retry=False):
        super(WebCacheUrlOpener, self).http_error_401(url, fp, errcode, errmsg, headers, data, True)

    def http_error_407(self, url, fp, errcode, errmsg, headers, data=None, retry=False):
        super(WebCacheUrlOpener, self).http_error_407(self, url, fp, errcode, errmsg, headers, data, True)

    def prompt_user_passwd(self, host, realm):
        return self.cntlr.internet_user_password(host, realm)
'''



'''
This is an example of a plug-in to both GUI menu and command line/web service
that will provide an option to replace behavior of table linkbase validation to
generate vs diff table linkbase infoset files.

See COPYRIGHT.md for copyright information.
'''
from arelle.Version import authorLabel, copyrightLabel

def validateTableInfosetMenuEntender(cntlr, validateMenu):
    # Extend menu with an item for the save infoset plugin
    cntlr.modelManager.generateTableInfoset = cntlr.config.setdefault("generateTableInfoset",False)
    from tkinter import BooleanVar
    generateTableInfoset = BooleanVar(value=cntlr.modelManager.generateTableInfoset)
    def setTableInfosetOption(*args):
        cntlr.config["generateTableInfoset"] = cntlr.modelManager.generateTableInfoset = generateTableInfoset.get()
    generateTableInfoset.trace("w", setTableInfosetOption)
    validateMenu.add_checkbutton(label=_("Generate table infosets (instead of diffing them)"),
                                 underline=0,
                                 variable=generateTableInfoset, onvalue=True, offvalue=False)

def validateTableInfosetCommandLineOptionExtender(parser):
    # extend command line options with a save DTS option
    parser.add_option("--generate-table-infoset",
                      action="store_true",
                      dest="generateTableInfoset",
                      help=_("Generate table instance infosets (instead of diffing them)."))

def validateTableInfosetCommandLineXbrlLoaded(cntlr, options, modelXbrl, *args, **kwargs):
    cntlr.modelManager.generateTableInfoset = getattr(options, "generateTableInfoset", False)

def validateTableInfoset(modelXbrl, resultTableUri):
    diffToFile = not getattr(modelXbrl.modelManager, 'generateTableInfoset', False)
    from arelle import ViewFileRenderedGrid
    ViewFileRenderedGrid.viewRenderedGrid(modelXbrl,
                                          resultTableUri,
                                          diffToFile=diffToFile)  # false to save infoset files
    return True # blocks standard behavior in validate.py

__pluginInfo__ = {
    'name': 'Validate Table Infoset (Optional behavior)',
    'version': '0.9',
    'description': "This plug-in adds a feature modify batch validation of table linkbase to save, versus diff, infoset files.  ",
    'license': 'Apache-2',
    'author': authorLabel,
    'copyright': copyrightLabel,
    # classes of mount points (required)
    'CntlrWinMain.Menu.Validation': validateTableInfosetMenuEntender,
    'CntlrCmdLine.Options': validateTableInfosetCommandLineOptionExtender,
    'CntlrCmdLine.Xbrl.Loaded': validateTableInfosetCommandLineXbrlLoaded,
    'Validate.TableInfoset': validateTableInfoset,
}
